from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING, Callable, Final

import pact
import pytest

from great_expectations.data_context import CloudDataContext
from great_expectations.exceptions import DataContextError
from tests.integration.cloud.rest_contracts.conftest import (
    EXISTING_ORGANIZATION_ID,
    ContractInteraction,
)

if TYPE_CHECKING:
    from requests import Session

    from tests.integration.cloud.rest_contracts.conftest import PactBody


NON_EXISTENT_EXPECTATION_SUITE_ID: Final[str] = "6ed9a340-8469-4ee2-a300-ffbe5d09b49d"
GET_EXPECTATION_SUITE_ID: Final[str] = "c138767f-1d62-4312-bfff-1167891ab76f"
PUT_EXPECTATION_SUITE_ID: Final[str] = "9390c24d-e8d6-4944-9411-4d0aaed14915"

POST_EXPECTATION_SUITE_MIN_REQUEST_BODY: Final[PactBody] = {
    "data": {
        "id": None,
        "meta": {"great_expectations_version": "0.13.23"},
        "expectations": [
            {
                "kwargs": {"max_value": 3, "min_value": 1},
                "meta": {},
                "expectation_type": "expect_table_row_count_to_be_between",
            },
        ],
        "name": "brand new suite",
        "organization_id": EXISTING_ORGANIZATION_ID,
    },
}

POST_EXPECTATION_SUITE_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": {
        "created_by_id": pact.Format().uuid,
        "organization_id": pact.Format().uuid,
        "name": "brand new suite",
        "expectations": [
            {
                "expectation_type": "expect_table_row_count_to_be_between",
                "id": pact.Format().uuid,
                "kwargs": {"max_value": 3, "min_value": 1},
                "meta": {},
            }
        ],
        "id": pact.Format().uuid,
        "meta": {"great_expectations_version": "0.13.23"},
    }
}

GET_EXPECTATION_SUITE_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": {
        "created_by_id": pact.Format().uuid,
        "organization_id": "0ccac18e-7631-4bdd-8a42-3c35cce574c6",
        "name": pact.Like("no_checkpoint_suite"),
        "expectations": [
            {
                "type": "expect_column_values_to_be_between",
                "id": pact.Format().uuid,
                "kwargs": {
                    "column": "passenger_count",
                    "max_value": 5,
                    "min_value": 0,
                    "mostly": 0.97,
                },
                "meta": {},
            }
        ],
        "id": GET_EXPECTATION_SUITE_ID,
        "meta": {"great_expectations_version": "0.18.3"},
    },
}

GET_EXPECTATION_SUITES_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": pact.EachLike(
        {
            "created_by_id": pact.Format().uuid,
            "organization_id": "0ccac18e-7631-4bdd-8a42-3c35cce574c6",
            "name": pact.Like("raw_health.critical_1a"),
            "id": pact.Format().uuid,
            "meta": {"great_expectations_version": pact.Like("0.13.23")},
            "expectations": pact.EachLike(
                {
                    "type": pact.Like("expect_column_values_to_be_between"),
                    "id": pact.Format().uuid,
                    "kwargs": pact.Like(
                        {
                            "column": "passenger_count",
                            "other_field": "another_value",
                        }
                    ),
                    "meta": pact.Like({}),
                }
            ),
        },
        minimum=1,
    ),
}


@pytest.mark.cloud
def test_get_expectation_suite(
    pact_test: pact.Pact,
    cloud_data_context: CloudDataContext,
    gx_cloud_session: Session,
) -> None:
    provider_state = "the Expectation Suite does exist"
    scenario = "a request to get an Expectation Suite"
    method = "GET"
    path = pathlib.Path(
        "/",
        "api",
        "v1",
        "organizations",
        EXISTING_ORGANIZATION_ID,
        "expectation-suites",
    )
    status = 200
    response_body = GET_EXPECTATION_SUITE_MIN_RESPONSE_BODY

    (
        pact_test.given(provider_state=provider_state)
        .upon_receiving(scenario=scenario)
        .with_request(
            method=method,
            path=str(path),
            headers=dict(gx_cloud_session.headers),
        )
        .will_respond_with(
            status=status,
            body=response_body,
        )
    )

    with pact_test:
        cloud_data_context.suites.get("no_checkpoint_suite")


@pytest.mark.cloud
def test_get_non_existent_expectation_suite(
    pact_test: pact.Pact,
    cloud_data_context: CloudDataContext,
    gx_cloud_session: Session,
) -> None:
    provider_state = "the Expectation Suite does not exist"
    scenario = "a request to get an Expectation Suite"
    method = "GET"
    path = pathlib.Path(
        "/",
        "api",
        "v1",
        "organizations",
        EXISTING_ORGANIZATION_ID,
        "expectation-suites",
    )

    status = 404

    (
        pact_test.given(provider_state=provider_state)
        .upon_receiving(scenario=scenario)
        .with_request(
            method=method,
            path=str(path),
            headers=dict(gx_cloud_session.headers),
        )
        .will_respond_with(
            status=status,
        )
    )

    with pact_test:
        with pytest.raises(DataContextError):
            cloud_data_context.suites.get(name="non_existent")


@pytest.mark.cloud
def test_get_expectation_suites(
    pact_test: pact.Pact,
    cloud_data_context: CloudDataContext,
    gx_cloud_session: Session,
) -> None:
    provider_state = "Expectation Suite exist"
    scenario = "a request to get Expectation Suites"
    method = "GET"
    path = pathlib.Path(
        "/",
        "api",
        "v1",
        "organizations",
        EXISTING_ORGANIZATION_ID,
        "expectation-suites",
    )
    status = 200
    response_body = GET_EXPECTATION_SUITES_MIN_RESPONSE_BODY

    (
        pact_test.given(provider_state=provider_state)
        .upon_receiving(scenario=scenario)
        .with_request(
            method=method,
            path=str(path),
            headers=dict(gx_cloud_session.headers),
        )
        .will_respond_with(
            status=status,
            body=response_body,
        )
    )

    with pact_test:
        cloud_data_context.suites.all()


@pytest.mark.cloud
@pytest.mark.parametrize(
    "contract_interaction",
    [
        ContractInteraction(
            method="POST",
            request_path=pathlib.Path(
                "/",
                "api",
                "v1",
                "organizations",
                EXISTING_ORGANIZATION_ID,
                "expectation-suites",
            ),
            upon_receiving="a request to post an Expectation Suite",
            given="the Expectation Suite does not exist",
            request_body={
                "data": {
                    "meta": {"great_expectations_version": "0.13.23"},
                    "expectations": [
                        {
                            "kwargs": {"max_value": 3, "min_value": 1},
                            "meta": {},
                            "expectation_type": "expect_table_row_count_to_be_between",
                        },
                    ],
                    "name": "brand new suite",
                },
            },
            response_status=201,
            response_body=POST_EXPECTATION_SUITE_MIN_RESPONSE_BODY,
        ),
    ],
)
def test_post_expectation_suite_request(
    contract_interaction: ContractInteraction,
    run_rest_api_pact_test: Callable[[ContractInteraction], None],
) -> None:
    run_rest_api_pact_test(contract_interaction)


@pytest.mark.cloud
@pytest.mark.parametrize(
    "contract_interaction",
    [
        ContractInteraction(
            method="PUT",
            request_path=pathlib.Path(
                "/",
                "api",
                "v1organizations",
                EXISTING_ORGANIZATION_ID,
                "expectation-suites",
                PUT_EXPECTATION_SUITE_ID,
            ),
            upon_receiving="a request to put an Expectation Suite",
            given="the Expectation Suite does exist",
            request_body={
                "data": {
                    "meta": {"great_expectations_version": "0.13.23"},
                    "expectations": [
                        {
                            "kwargs": {"max_value": 3, "min_value": 1},
                            "meta": {},
                            "expectation_type": "expect_table_row_count_to_be_between",
                        },
                    ],
                    "name": "renamed suite",
                },
            },
            response_status=200,
            response_body=None,
        ),
    ],
)
def test_put_expectation_suite_request(
    contract_interaction: ContractInteraction,
    run_rest_api_pact_test: Callable[[ContractInteraction], None],
) -> None:
    run_rest_api_pact_test(contract_interaction)


@pytest.mark.cloud
@pytest.mark.parametrize(
    "contract_interaction",
    [
        ContractInteraction(
            method="PUT",
            request_path=pathlib.Path(
                "/",
                "api",
                "v1",
                "organizations",
                EXISTING_ORGANIZATION_ID,
                "expectation-suites",
                NON_EXISTENT_EXPECTATION_SUITE_ID,
            ),
            upon_receiving="a request to put an Expectation Suite",
            given="the Expectation Suite does not exist",
            request_body={
                "data": {
                    "meta": {"great_expectations_version": "0.13.23"},
                    "expectations": [
                        {
                            "kwargs": {"max_value": 3, "min_value": 1},
                            "meta": {},
                            "expectation_type": "expect_table_row_count_to_be_between",
                        },
                    ],
                    "name": "renamed suite",
                },
            },
            response_status=404,
            response_body=None,
        ),
    ],
)
def test_put_non_existent_expectation_suite(
    contract_interaction: ContractInteraction,
    run_rest_api_pact_test: Callable[[ContractInteraction], None],
) -> None:
    run_rest_api_pact_test(contract_interaction)



import pytest

from great_expectations.core.expectation_diagnostics.supporting_types import Maturity
from tests.expectations.fixtures.expect_column_values_to_equal_three import (
    ExpectColumnValuesToEqualThree,
    ExpectColumnValuesToEqualThree__SecondIteration,
    ExpectColumnValuesToEqualThree__ThirdIteration,
)

# module level markers
pytestmark = pytest.mark.unit


@pytest.mark.skip("This is broken because Expectation._get_execution_engine_diagnostics is broken")
def test_print_diagnostic_checklist__first_iteration():
    output_message = ExpectColumnValuesToEqualThree().print_diagnostic_checklist()

    assert (
        output_message
        == """\
Completeness checklist for ExpectColumnValuesToEqualThree:
   library_metadata object exists
   Has a docstring, including a one-line short description that begins with "Expect" and ends with a period
   Has at least one positive and negative example case, and all test cases pass
   Has core logic and passes tests on at least one Execution Engine
"""  # noqa: E501 # FIXME CoP
    )


def test_print_diagnostic_checklist__second_iteration():
    output_message = ExpectColumnValuesToEqualThree__SecondIteration(
        column="values"
    ).print_diagnostic_checklist()
    print(output_message)

    assert (
        output_message
        == f"""\
Completeness checklist for ExpectColumnValuesToEqualThree__SecondIteration ({Maturity.EXPERIMENTAL}):
 ✔ Has a valid library_metadata object
 ✔ Has a docstring, including a one-line short description that begins with "Expect" and ends with a period
    ✔ "Expect values in this column to equal the number three."
 ✔ Has at least one positive and negative example case, and all test cases pass
 ✔ Has core logic and passes tests on at least one Execution Engine
    ✔ All 3 tests for pandas are passing
   Has basic input validation and type checking
      No validate_configuration method defined on subclass
 ✔ Has both statement Renderers: prescriptive and diagnostic
 ✔ Has core logic that passes tests for all applicable Execution Engines and SQL dialects
    ✔ All 3 tests for pandas are passing
   Has a full suite of tests, as determined by a code owner
   Has passed a manual review by a code owner for code standards and style guides
"""  # noqa: E501 # FIXME CoP
    )


def test_print_diagnostic_checklist__third_iteration():
    output_message = ExpectColumnValuesToEqualThree__ThirdIteration(
        column="values"
    ).print_diagnostic_checklist()
    print(output_message)

    assert (
        output_message
        == f"""\
Completeness checklist for ExpectColumnValuesToEqualThree__ThirdIteration ({Maturity.EXPERIMENTAL}):
 ✔ Has a valid library_metadata object
   Has a docstring, including a one-line short description that begins with "Expect" and ends with a period
 ✔ Has at least one positive and negative example case, and all test cases pass
 ✔ Has core logic and passes tests on at least one Execution Engine
    ✔ All 3 tests for pandas are passing
   Has basic input validation and type checking
      No validate_configuration method defined on subclass
 ✔ Has both statement Renderers: prescriptive and diagnostic
 ✔ Has core logic that passes tests for all applicable Execution Engines and SQL dialects
    ✔ All 3 tests for pandas are passing
   Has a full suite of tests, as determined by a code owner
   Has passed a manual review by a code owner for code standards and style guides
"""  # noqa: E501 # FIXME CoP
    )



from __future__ import annotations

import configparser
import os
import pathlib
import shutil
import uuid
from pathlib import Path
from typing import Dict, List, Union

import pytest

import great_expectations as gx
import great_expectations.exceptions as gx_exceptions
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import get_context
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.data_context.data_context.cloud_data_context import CloudDataContext
from great_expectations.data_context.data_context.ephemeral_data_context import (
    EphemeralDataContext,
)
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.store import ExpectationsStore
from great_expectations.data_context.util import file_relative_path
from great_expectations.expectations.expectation import BatchExpectation
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.render import (
    AtomicPrescriptiveRendererType,
    AtomicRendererType,
    RenderedAtomicContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.util import (
    gen_directory_tree_str,
)
from tests.test_utils import safe_remove

yaml = YAMLHandler()

parameterized_expectation_suite_name = "my_dag_node.default"


@pytest.fixture
def data_context_with_bad_datasource(tmp_path_factory):
    """
    This data_context is *manually* created to have the config we want.

    This DataContext has a connection to a datasource named my_postgres_db
    which is not a valid datasource.

    It is used by test_get_batch_multiple_datasources_do_not_scan_all()
    """
    project_path = str(tmp_path_factory.mktemp("data_context"))
    context_path = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    asset_config_path = os.path.join(context_path, "expectations")  # noqa: PTH118 # FIXME CoP
    fixture_dir = file_relative_path(__file__, "../test_fixtures")
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(asset_config_path, "my_dag_node"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    shutil.copy(
        os.path.join(  # noqa: PTH118 # FIXME CoP
            fixture_dir, "great_expectations_bad_datasource.yml"
        ),
        str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
    )
    return get_context(context_root_dir=context_path)


@pytest.mark.cloud
def test_get_expectation_suite_include_rendered_content(
    empty_cloud_data_context: CloudDataContext,
):
    context = empty_cloud_data_context

    expectation_suite: ExpectationSuite = context.suites.add(
        ExpectationSuite(name="this_data_asset_config_does_not_exist.default")
    )
    expectation_suite.expectation_configurations.append(
        ExpectationConfiguration(type="expect_table_row_count_to_equal", kwargs={"value": 10})
    )
    for expectation in expectation_suite.expectation_configurations:
        assert expectation.rendered_content is None
    expectation_suite.save()
    context.suites.get("this_data_asset_config_does_not_exist.default")
    for expectation in expectation_suite.expectation_configurations:
        assert expectation.rendered_content is None

    expectation_suite_retrieved: ExpectationSuite = context.suites.get(
        "this_data_asset_config_does_not_exist.default",
    )

    for expectation in expectation_suite_retrieved.expectation_configurations:
        assert expectation.rendered_content
        for rendered_content_block in expectation.rendered_content:
            assert isinstance(
                rendered_content_block,
                RenderedAtomicContent,
            )


@pytest.mark.unit
def test_data_context_get_datasource_on_non_existent_one_raises_helpful_error(
    titanic_data_context: AbstractDataContext,
):
    # this is deprecated
    with pytest.warns(DeprecationWarning), pytest.raises(ValueError):
        _ = titanic_data_context.get_datasource("fakey_mc_fake")


@pytest.mark.unit
def test_data_context_get_datasource(
    titanic_data_context: AbstractDataContext,
):
    # this is deprecated
    name = "my datasource"
    ds = titanic_data_context.data_sources.add_pandas(name)
    with pytest.warns(DeprecationWarning) as warning_records:
        fetched_ds = titanic_data_context.get_datasource(name)

    assert fetched_ds == ds
    assert len(warning_records) == 1
    assert "context.get_datasource is deprecated" in str(warning_records.list[0].message)


@pytest.mark.unit
def test_add_store(empty_data_context):
    assert "my_new_store" not in empty_data_context.stores
    assert "my_new_store" not in empty_data_context.get_config()["stores"]
    new_store = empty_data_context.add_store(
        "my_new_store",
        {
            "module_name": "great_expectations.data_context.store",
            "class_name": "ExpectationsStore",
        },
    )
    assert "my_new_store" in empty_data_context.stores
    assert "my_new_store" in empty_data_context.get_config()["stores"]

    assert isinstance(new_store, ExpectationsStore)


@pytest.mark.unit
def test_ConfigOnlyDataContext__initialization(tmp_path_factory, basic_data_context_config):
    config_path = str(tmp_path_factory.mktemp("test_ConfigOnlyDataContext__initialization__dir"))
    context = get_context(
        basic_data_context_config,
        config_path,
    )

    assert context.root_directory.split("/")[-1].startswith(
        "test_ConfigOnlyDataContext__initialization__dir"
    )

    plugins_dir_parts = context.plugins_directory.split("/")[-3:]
    assert len(plugins_dir_parts) == 3
    assert plugins_dir_parts[0].startswith("test_ConfigOnlyDataContext__initialization__dir")
    assert plugins_dir_parts[1:] == ["plugins", ""]


@pytest.mark.unit
def test__normalize_absolute_or_relative_path(tmp_path_factory, basic_data_context_config):
    full_test_dir = tmp_path_factory.mktemp("test__normalize_absolute_or_relative_path__dir")
    test_dir = full_test_dir.parts[-1]
    config_path = str(full_test_dir)
    context = get_context(
        basic_data_context_config,
        config_path,
    )

    assert context._normalize_absolute_or_relative_path("yikes").endswith(
        os.path.join(test_dir, "yikes")  # noqa: PTH118 # FIXME CoP
    )

    assert test_dir not in context._normalize_absolute_or_relative_path("/yikes")
    assert context._normalize_absolute_or_relative_path("/yikes") == "/yikes"


@pytest.mark.filesystem
def test_load_data_context_from_environment_variables(tmp_path, monkeypatch):
    # `find_context_root_dir` iterates up the file tree to find a great_expectations.yml
    # By deeply nesting our project path, we ensure we don't collide with any existing
    # fixtures or side effects from other tests
    project_path = tmp_path / "a" / "b" / "c" / "d" / "data_context"
    project_path.mkdir(parents=True)

    context_path = project_path / FileDataContext.GX_DIR
    context_path.mkdir()
    monkeypatch.chdir(str(context_path))

    with pytest.raises(gx_exceptions.ConfigNotFoundError):
        FileDataContext.find_context_root_dir()

    shutil.copy(
        file_relative_path(
            __file__,
            os.path.join(  # noqa: PTH118 # FIXME CoP
                "..", "test_fixtures", "great_expectations_basic.yml"
            ),
        ),
        str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
    )
    monkeypatch.setenv("GX_HOME", str(context_path))
    assert FileDataContext.find_context_root_dir() == str(context_path)


@pytest.mark.filesystem
def test_data_context_create_does_not_raise_error_or_warning_if_ge_dir_exists(
    tmp_path_factory,
):
    project_path = str(tmp_path_factory.mktemp("data_context"))
    gx.get_context(mode="file", project_root_dir=project_path)


@pytest.fixture()
def empty_context(tmp_path_factory) -> FileDataContext:
    project_path = str(tmp_path_factory.mktemp("data_context"))
    ge_dir = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    context = get_context(context_root_dir=ge_dir)
    assert isinstance(context, FileDataContext)
    assert os.path.isdir(ge_dir)  # noqa: PTH112 # FIXME CoP
    assert os.path.isfile(  # noqa: PTH113 # FIXME CoP
        os.path.join(ge_dir, FileDataContext.GX_YML)  # noqa: PTH118 # FIXME CoP
    )
    return context


@pytest.mark.filesystem
def test_data_context_is_project_scaffolded(empty_context):
    ge_dir = empty_context.root_directory
    assert FileDataContext.is_project_scaffolded(ge_dir) is True


@pytest.mark.filesystem
def test_data_context_does_ge_yml_exist_returns_true_when_it_does_exist(empty_context):
    ge_dir = empty_context.root_directory
    assert FileDataContext.does_config_exist_on_disk(ge_dir) is True


@pytest.mark.filesystem
def test_data_context_does_ge_yml_exist_returns_false_when_it_does_not_exist(
    empty_context,
):
    ge_dir = empty_context.root_directory
    # mangle project
    safe_remove(os.path.join(ge_dir, empty_context.GX_YML))  # noqa: PTH118 # FIXME CoP
    assert FileDataContext.does_config_exist_on_disk(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_does_project_have_a_datasource_in_config_file_returns_true_when_it_has_a_datasource_configured_in_yml_file_on_disk(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    empty_context.data_sources.add_pandas("arthur")
    assert FileDataContext._does_project_have_a_datasource_in_config_file(ge_dir) is True


@pytest.mark.filesystem
def test_data_context_does_project_have_a_datasource_in_config_file_returns_false_when_it_does_not_have_a_datasource_configured_in_yml_file_on_disk(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    assert FileDataContext._does_project_have_a_datasource_in_config_file(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_does_project_have_a_datasource_in_config_file_returns_false_when_it_does_not_have_a_ge_yml_file(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    safe_remove(os.path.join(ge_dir, empty_context.GX_YML))  # noqa: PTH118 # FIXME CoP
    assert FileDataContext._does_project_have_a_datasource_in_config_file(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_does_project_have_a_datasource_in_config_file_returns_false_when_it_does_not_have_a_ge_dir(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    safe_remove(os.path.join(ge_dir))  # noqa: PTH118 # FIXME CoP
    assert FileDataContext._does_project_have_a_datasource_in_config_file(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_does_project_have_a_datasource_in_config_file_returns_false_when_the_project_has_an_invalid_config_file(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    with open(os.path.join(ge_dir, FileDataContext.GX_YML), "w") as yml:  # noqa: PTH118 # FIXME CoP
        yml.write("this file: is not a valid ge config")
    assert FileDataContext._does_project_have_a_datasource_in_config_file(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_true_when_its_valid_context_has_one_datasource_and_one_suite(  # noqa: E501 # FIXME CoP
    empty_context,
):
    context = empty_context
    ge_dir = context.root_directory
    context.data_sources.add_pandas("arthur")
    context.suites.add(ExpectationSuite(name="dent"))
    assert len(context.suites.all()) == 1

    assert FileDataContext.is_project_initialized(ge_dir) is True


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_true_when_its_valid_context_has_one_datasource_and_no_suites(  # noqa: E501 # FIXME CoP
    empty_context,
):
    context = empty_context
    ge_dir = context.root_directory
    context.data_sources.add_pandas("arthur")
    assert len(context.suites.all()) == 0

    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_false_when_its_valid_context_has_no_datasource(
    empty_context,
):
    ge_dir = empty_context.root_directory
    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_false_when_config_yml_is_missing(
    empty_context,
):
    ge_dir = empty_context.root_directory
    # mangle project
    safe_remove(os.path.join(ge_dir, empty_context.GX_YML))  # noqa: PTH118 # FIXME CoP

    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_false_when_uncommitted_dir_is_missing(
    empty_context,
):
    ge_dir = empty_context.root_directory
    # mangle project
    shutil.rmtree(
        os.path.join(ge_dir, empty_context.GX_UNCOMMITTED_DIR)  # noqa: PTH118 # FIXME CoP
    )

    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_false_when_uncommitted_data_docs_dir_is_missing(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    # mangle project
    shutil.rmtree(
        os.path.join(  # noqa: PTH118 # FIXME CoP
            ge_dir, empty_context.GX_UNCOMMITTED_DIR, "data_docs"
        )
    )

    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_false_when_uncommitted_validations_dir_is_missing(  # noqa: E501 # FIXME CoP
    empty_context,
):
    ge_dir = empty_context.root_directory
    # mangle project
    shutil.rmtree(
        os.path.join(  # noqa: PTH118 # FIXME CoP
            ge_dir, empty_context.GX_UNCOMMITTED_DIR, "validations"
        )
    )

    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_is_project_initialized_returns_false_when_config_variable_yml_is_missing(
    empty_context,
):
    ge_dir = empty_context.root_directory
    # mangle project
    safe_remove(
        os.path.join(  # noqa: PTH118 # FIXME CoP
            ge_dir, empty_context.GX_UNCOMMITTED_DIR, "config_variables.yml"
        )
    )

    assert FileDataContext.is_project_initialized(ge_dir) is False


@pytest.mark.filesystem
def test_data_context_create_raises_warning_and_leaves_existing_yml_untouched(
    tmp_path_factory,
):
    project_path = str(tmp_path_factory.mktemp("data_context"))
    gx.get_context(mode="file", project_root_dir=project_path)
    ge_yml = os.path.join(project_path, "gx/great_expectations.yml")  # noqa: PTH118 # FIXME CoP
    with open(ge_yml, "a") as ff:
        ff.write("# LOOK I WAS MODIFIED")

    gx.get_context(mode="file", project_root_dir=project_path)

    with open(ge_yml) as ff:
        obs = ff.read()
    assert "# LOOK I WAS MODIFIED" in obs


@pytest.mark.filesystem
def test_data_context_create_makes_uncommitted_dirs_when_all_are_missing(
    tmp_path_factory,
):
    project_path = str(tmp_path_factory.mktemp("data_context"))
    gx.get_context(mode="file", project_root_dir=project_path)

    # mangle the existing setup
    ge_dir = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    uncommitted_dir = os.path.join(ge_dir, "uncommitted")  # noqa: PTH118 # FIXME CoP
    shutil.rmtree(uncommitted_dir)

    # re-run create to simulate onboarding
    gx.get_context(mode="file", project_root_dir=project_path)
    obs = gen_directory_tree_str(ge_dir)

    assert os.path.isdir(  # noqa: PTH112 # FIXME CoP
        uncommitted_dir
    ), "No uncommitted directory created"
    assert (
        obs
        == """\
gx/
    .gitignore
    great_expectations.yml
    checkpoints/
    expectations/
        .ge_store_backend_id
    plugins/
        custom_data_docs/
            renderers/
            styles/
                data_docs_custom_styles.css
            views/
    uncommitted/
        config_variables.yml
        data_docs/
        validations/
            .ge_store_backend_id
    validation_definitions/
"""
    )


@pytest.mark.filesystem
def test_data_context_create_does_nothing_if_all_uncommitted_dirs_exist(
    tmp_path_factory,
):
    expected = """\
gx/
    .gitignore
    great_expectations.yml
    checkpoints/
    expectations/
        .ge_store_backend_id
    plugins/
        custom_data_docs/
            renderers/
            styles/
                data_docs_custom_styles.css
            views/
    uncommitted/
        config_variables.yml
        data_docs/
        validations/
            .ge_store_backend_id
    validation_definitions/
"""
    project_path = str(tmp_path_factory.mktemp("stuff"))
    ge_dir = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP

    gx.get_context(mode="file", project_root_dir=project_path)
    fixture = gen_directory_tree_str(ge_dir)

    assert fixture == expected

    # re-run create to simulate onboarding
    gx.get_context(mode="file", project_root_dir=project_path)

    obs = gen_directory_tree_str(ge_dir)
    assert obs == expected


@pytest.mark.filesystem
def test_data_context_do_all_uncommitted_dirs_exist(tmp_path_factory):
    expected = """\
uncommitted/
    config_variables.yml
    data_docs/
    validations/
        .ge_store_backend_id
"""
    project_path = str(tmp_path_factory.mktemp("stuff"))
    ge_dir = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    uncommitted_dir = os.path.join(ge_dir, "uncommitted")  # noqa: PTH118 # FIXME CoP
    gx.get_context(mode="file", project_root_dir=project_path)
    fixture = gen_directory_tree_str(uncommitted_dir)
    assert fixture == expected

    # Test that all exist
    assert FileDataContext.all_uncommitted_directories_exist(ge_dir)

    # remove a few
    shutil.rmtree(os.path.join(uncommitted_dir, "data_docs"))  # noqa: PTH118 # FIXME CoP
    shutil.rmtree(os.path.join(uncommitted_dir, "validations"))  # noqa: PTH118 # FIXME CoP

    # Test that not all exist
    assert not FileDataContext.all_uncommitted_directories_exist(project_path)


@pytest.mark.filesystem
def test_data_context_create_builds_base_directories(tmp_path_factory):
    project_path = str(tmp_path_factory.mktemp("data_context"))
    context = gx.get_context(mode="file", project_root_dir=project_path)
    assert isinstance(context, FileDataContext)

    for directory in [
        "expectations",
        "plugins",
        "checkpoints",
        "uncommitted",
    ]:
        base_dir = os.path.join(project_path, context.GX_DIR, directory)  # noqa: PTH118 # FIXME CoP
        assert os.path.isdir(base_dir)  # noqa: PTH112 # FIXME CoP


@pytest.mark.filesystem
def test_data_context_create_does_not_overwrite_existing_config_variables_yml(
    tmp_path_factory,
):
    project_path = str(tmp_path_factory.mktemp("data_context"))
    gx.get_context(mode="file", project_root_dir=project_path)
    ge_dir = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    uncommitted_dir = os.path.join(ge_dir, "uncommitted")  # noqa: PTH118 # FIXME CoP
    config_vars_yml = os.path.join(  # noqa: PTH118 # FIXME CoP
        uncommitted_dir, "config_variables.yml"
    )

    # modify config variables
    with open(config_vars_yml, "a") as ff:
        ff.write("# LOOK I WAS MODIFIED")

    # re-run create to simulate onboarding
    gx.get_context(mode="file", project_root_dir=project_path)

    with open(config_vars_yml) as ff:
        obs = ff.read()
    assert "# LOOK I WAS MODIFIED" in obs


@pytest.mark.filesystem
def test_scaffold_directories(tmp_path_factory):
    empty_directory = str(tmp_path_factory.mktemp("test_scaffold_directories"))
    empty_directory_path = Path(empty_directory)
    FileDataContext._scaffold_directories(empty_directory_path)

    assert {p.name for p in empty_directory_path.iterdir()} == {
        "plugins",
        "checkpoints",
        "expectations",
        ".gitignore",
        "uncommitted",
        "validation_definitions",
    }
    assert {p.name for p in (empty_directory_path / "uncommitted").iterdir()} == {
        "data_docs",
        "validations",
    }


@pytest.mark.filesystem
def test_load_config_variables_property(basic_data_context_config, tmp_path_factory, monkeypatch):
    # Setup:
    base_path = str(tmp_path_factory.mktemp("test_load_config_variables_file"))
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(base_path, "uncommitted"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    with open(
        os.path.join(base_path, "uncommitted", "dev_variables.yml"),  # noqa: PTH118 # FIXME CoP
        "w",
    ) as outfile:
        yaml.dump({"env": "dev"}, outfile)
    with open(
        os.path.join(base_path, "uncommitted", "prod_variables.yml"),  # noqa: PTH118 # FIXME CoP
        "w",
    ) as outfile:
        yaml.dump({"env": "prod"}, outfile)
    basic_data_context_config["config_variables_file_path"] = (
        "uncommitted/${TEST_CONFIG_FILE_ENV}_variables.yml"
    )

    try:
        # We should be able to load different files based on an environment variable
        monkeypatch.setenv("TEST_CONFIG_FILE_ENV", "dev")
        context = get_context(basic_data_context_config, context_root_dir=base_path)
        config_vars = context.config_variables
        assert config_vars["env"] == "dev"
        monkeypatch.setenv("TEST_CONFIG_FILE_ENV", "prod")
        context = get_context(basic_data_context_config, context_root_dir=base_path)
        config_vars = context.config_variables
        assert config_vars["env"] == "prod"
    finally:
        # Make sure we unset the environment variable we're using
        monkeypatch.delenv("TEST_CONFIG_FILE_ENV")


@pytest.mark.filesystem
def test_add_expectation_to_expectation_suite(empty_data_context_stats_enabled):
    context = empty_data_context_stats_enabled

    expectation_suite: ExpectationSuite = context.suites.add(
        ExpectationSuite(name="my_new_expectation_suite")
    )
    expectation_suite.add_expectation_configuration(
        ExpectationConfiguration(type="expect_table_row_count_to_equal", kwargs={"value": 10})
    )


@pytest.mark.filesystem
def test_modifications_to_env_vars_is_recognized_within_same_program_execution(
    empty_data_context, monkeypatch
) -> None:
    """
    What does this test do and why?

    Great Expectations recognizes changes made to environment variables within a program execution
    and ensures that these changes are recognized in subsequent calls within the same process.

    This is particularly relevant when performing substitutions within a user's project config.
    """
    context = empty_data_context
    env_var_name: str = "MY_PLUGINS_DIRECTORY"
    env_var_value: str = "my_patched_value"

    context.variables.config.plugins_directory = f"${env_var_name}"
    monkeypatch.setenv(env_var_name, env_var_value)

    assert context.plugins_directory and context.plugins_directory.endswith(env_var_value)


@pytest.mark.filesystem
def test_modifications_to_config_vars_is_recognized_within_same_program_execution(
    empty_data_context,
) -> None:
    """
    What does this test do and why?

    Great Expectations recognizes changes made to config variables within a program execution
    and ensures that these changes are recognized in subsequent calls within the same process.

    This is particularly relevant when performing substitutions within a user's project config.
    """
    context = empty_data_context
    config_var_name: str = "my_plugins_dir"
    config_var_value: str = "my_patched_value"

    context.variables.config.plugins_directory = f"${config_var_name}"
    context.save_config_variable(name=config_var_name, value=config_var_value)

    assert context.plugins_directory and context.plugins_directory.endswith(config_var_value)


class ExpectSkyToBeColor(BatchExpectation):
    metric_dependencies = ("table.color",)
    success_keys = ("color",)
    args_keys = ("color",)

    color: str

    @classmethod
    @renderer(renderer_type=".".join([AtomicRendererType.PRESCRIPTIVE, "custom_renderer_type"]))
    def _prescriptive_renderer_custom(
        cls,
        **kwargs: dict,
    ) -> None:
        raise ValueError("This renderer is broken!")

    def _validate(  # type: ignore[override,explicit-override] # FIXME
        self,
        **kwargs: dict,
    ) -> Dict[str, Union[bool, dict]]:
        return {
            "success": True,
            "result": {"observed_value": "blue"},
        }


class TestRenderedContent:
    @pytest.mark.filesystem
    def test_no_rendered_content_for_file_data_context(self, empty_data_context: FileDataContext):
        suite_name = "test_suite"
        empty_data_context.suites.add(
            ExpectationSuite(
                name=suite_name,
                expectations=[gx.expectations.ExpectTableRowCountToEqual(value=0)],
            )
        )
        expectation_suite = empty_data_context.suites.get(name=suite_name)

        assert all(
            expectation.rendered_content is None for expectation in expectation_suite.expectations
        )

    @pytest.mark.cloud
    def test_rendered_content_for_cloud(self, empty_cloud_data_context: CloudDataContext) -> None:
        suite_name = "test_suite"
        empty_cloud_data_context.suites.add(
            ExpectationSuite(
                name=suite_name,
                expectations=[gx.expectations.ExpectTableRowCountToEqual(value=0)],
            )
        )
        expectation_suite = empty_cloud_data_context.suites.get(name=suite_name)

        rendered_content_blocks: list = []
        for expectation in expectation_suite.expectations:
            assert expectation.rendered_content is not None
            rendered_content_blocks.extend(expectation.rendered_content)
        assert rendered_content_blocks

    @pytest.mark.cloud
    def test_multiple_rendered_content_blocks_one_is_busted(
        self, empty_cloud_data_context: CloudDataContext
    ) -> None:
        suite_name = "test_suite"
        empty_cloud_data_context.suites.add(
            ExpectationSuite(
                name=suite_name,
                expectations=[
                    ExpectSkyToBeColor(color="blue"),
                ],
            )
        )
        expectation_suite = empty_cloud_data_context.suites.get(name=suite_name)

        expected_rendered_content: List[RenderedAtomicContent] = [
            RenderedAtomicContent(
                name=AtomicPrescriptiveRendererType.FAILED,
                value=renderedAtomicValueSchema.load(
                    {
                        "template": (
                            "Rendering failed for Expectation: $expectation_type(**$kwargs)."
                        ),
                        "params": {
                            "expectation_type": {
                                "schema": {"type": "string"},
                                "value": "expect_sky_to_be_color",
                            },
                            "kwargs": {
                                "schema": {"type": "string"},
                                "value": {"color": "blue"},
                            },
                        },
                        "schema": {"type": "com.superconductive.rendered.string"},
                    }
                ),
                value_type="StringValueType",
                exception='Renderer "atomic.prescriptive.custom_renderer_type" failed to render Expectation '  # noqa: E501 # FIXME CoP
                '"expect_sky_to_be_color with exception message: This renderer is broken!".',
            ),
            RenderedAtomicContent(
                name=AtomicPrescriptiveRendererType.SUMMARY,
                value=renderedAtomicValueSchema.load(
                    {
                        "schema": {"type": "com.superconductive.rendered.string"},
                        "template": "$expectation_type(**$kwargs)",
                        "params": {
                            "expectation_type": {
                                "schema": {"type": "string"},
                                "value": "expect_sky_to_be_color",
                            },
                            "kwargs": {
                                "schema": {"type": "string"},
                                "value": {"color": "blue"},
                            },
                        },
                    }
                ),
                value_type="StringValueType",
            ),
        ]

        actual_rendered_content: List[RenderedAtomicContent] = []
        for expectation in expectation_suite.expectations:
            assert expectation.rendered_content is not None
            actual_rendered_content.extend(expectation.rendered_content)

        assert actual_rendered_content == expected_rendered_content


@pytest.mark.filesystem
def test_file_backed_context_scaffolds_gitignore(tmp_path: pathlib.Path):
    project_path = tmp_path / "my_project_root_dir"
    context_path = project_path / FileDataContext.GX_DIR
    uncommitted = context_path / FileDataContext.GX_UNCOMMITTED_DIR
    gitignore = context_path / FileDataContext.GITIGNORE

    assert not uncommitted.exists()
    assert not gitignore.exists()

    # Scaffold project directory
    _ = get_context(context_root_dir=context_path)

    assert uncommitted.exists()
    assert gitignore.exists()
    assert FileDataContext.GX_UNCOMMITTED_DIR in gitignore.read_text()


@pytest.mark.filesystem
def test_file_backed_context_updates_existing_gitignore(tmp_path: pathlib.Path):
    project_path = tmp_path / "my_project_root_dir"
    context_path = project_path / FileDataContext.GX_DIR
    uncommitted = context_path / FileDataContext.GX_UNCOMMITTED_DIR
    gitignore = context_path / FileDataContext.GITIGNORE

    # Scaffold necessary files so `get_context` updates rather than creates
    uncommitted.mkdir(parents=True)
    existing_value = "__pycache__/"
    with gitignore.open("w") as f:
        f.write(f"\n{existing_value}")

    # Scaffold project directory
    _ = get_context(context_root_dir=context_path)

    contents = gitignore.read_text()
    assert existing_value in contents
    assert FileDataContext.GX_UNCOMMITTED_DIR in contents


@pytest.mark.unit
def test_set_oss_id_with_empty_config(in_memory_runtime_context: EphemeralDataContext):
    context = in_memory_runtime_context
    config = configparser.ConfigParser()

    oss_id = context._set_oss_id(config)

    assert config.sections() == ["analytics"]
    assert list(config["analytics"]) == ["oss_id"]
    assert oss_id == uuid.UUID(config["analytics"]["oss_id"])


@pytest.mark.unit
def test_set_oss_id_with_existing_config(
    in_memory_runtime_context: EphemeralDataContext,
):
    context = in_memory_runtime_context

    # Set up existing config
    config = configparser.ConfigParser()
    config["analytics"] = {}

    oss_id = context._set_oss_id(config)

    assert config.sections() == ["analytics"]
    assert list(config["analytics"]) == [
        "oss_id",
    ]
    assert oss_id == uuid.UUID(config["analytics"]["oss_id"])



from typing import Any, Dict, Generic, Optional

from typing_extensions import TypeAlias

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import StrictStr
from great_expectations.core.batch_definition import PartitionerT
from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice

BatchParameters: TypeAlias = Dict[StrictStr, Any]

class BatchRequest(pydantic.GenericModel, Generic[PartitionerT]):
    datasource_name: StrictStr
    data_asset_name: StrictStr
    options: BatchParameters
    partitioner: Optional[PartitionerT] = None

    def __init__(
        self,
        datasource_name: StrictStr,
        data_asset_name: StrictStr,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[PartitionerT] = None,
    ) -> None: ...
    @property
    def batch_slice(self) -> slice: ...

    # MyPy doesn't like setters/getters with different types (for now)
    # Type ignores can be avoided by using BatchRequest.update_batch_slice().
    # https://github.com/python/mypy/issues/3004
    @batch_slice.setter
    def batch_slice(self, value: Optional[BatchSlice]) -> None: ...
    def update_batch_slice(self, value: Optional[BatchSlice] = None) -> None: ...



from __future__ import annotations

from typing import Literal, Optional, Union

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_asset.path.directory_asset import (
    DirectoryDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.data_asset.path.spark.spark_generic import (
    _SparkGenericFilePathAssetMixin,
)


class ParquetAssetBase(_SparkGenericFilePathAssetMixin):
    # The options below are available as of spark v3.4.0
    # See https://spark.apache.org/docs/latest/sql-data-sources-parquet.html for more info.
    merge_schema: Optional[Union[bool, str]] = Field(None, alias="mergeSchema")
    datetime_rebase_mode: Optional[Literal["EXCEPTION", "CORRECTED", "LEGACY"]] = Field(
        None, alias="datetimeRebaseMode"
    )
    int_96_rebase_mode: Optional[Literal["EXCEPTION", "CORRECTED", "LEGACY"]] = Field(
        None, alias="int96RebaseMode"
    )

    class Config:
        extra = "forbid"

        allow_population_by_field_name = True

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "parquet"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-parquet.html for more info.
        """
        return (
            super()
            ._get_reader_options_include()
            .union(
                {
                    "datetime_rebase_mode",
                    "int_96_rebase_mode",
                    "merge_schema",
                }
            )
        )


class ParquetAsset(FileDataAsset, ParquetAssetBase):
    type: Literal["parquet"] = "parquet"


class DirectoryParquetAsset(DirectoryDataAsset, ParquetAssetBase):
    type: Literal["directory_parquet"] = "directory_parquet"

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "parquet"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-parquet.html for more info.
        """
        return (
            super()._get_reader_options_include()
            | super(DirectoryDataAsset, self)._get_reader_options_include()
        )



import pytest

pytestmarks = pytest.mark.filesystem


def read_config_from_file(config_filename):
    with open(config_filename) as f_:
        config = f_.read()

    return config


@pytest.mark.filesystem
def test_add_store_immediately_adds_to_config(empty_data_context):
    context = empty_data_context
    config_filename = context.root_directory + "/great_expectations.yml"

    assert "my_new_store" not in read_config_from_file(config_filename)
    context.add_store(
        "my_new_store",
        {
            "module_name": "great_expectations.data_context.store",
            "class_name": "ExpectationsStore",
        },
    )
    assert "my_new_store" in read_config_from_file(config_filename)



# Utility methods for dealing with DataAsset objects
from __future__ import annotations

import datetime
import decimal
import sys
from typing import Any

import numpy as np
import pandas as pd

from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.types import SerializableDictDot, SerializableDotDict


def recursively_convert_to_json_serializable(
    test_obj: dict,
) -> dict:
    """
    Helper function to convert an object to one that is serializable

    Args:
        test_obj: an object to attempt to convert a corresponding json-serializable object

    Returns:
        (dict) A converted test_object

    Warning:
        test_obj may also be converted in place.

    """
    return _recursively_convert_to_json_serializable(test_obj)


def _recursively_convert_to_json_serializable(  # noqa: C901, PLR0911, PLR0912 # FIXME CoP
    test_obj: Any,
) -> Any:
    # If it's one of our types, we pass
    if isinstance(test_obj, (SerializableDictDot, SerializableDotDict)):
        return test_obj

    # Validate that all aruguments are of approved types, coerce if it's easy, else exception
    # print(type(test_obj), test_obj)
    # Note: Not 100% sure I've resolved this correctly...
    try:
        if not isinstance(test_obj, list) and np.isnan(test_obj):
            # np.isnan is functionally vectorized, but we only want to apply this to single objects
            # Hence, why we test for `not isinstance(list)`
            return None
    except (TypeError, ValueError):
        pass

    if isinstance(test_obj, (str, int, float, bool)):
        # No problem to encode json
        return test_obj

    elif isinstance(test_obj, dict):
        new_dict = {}
        for key in test_obj:
            if key == "row_condition" and test_obj[key] is not None:
                ensure_row_condition_is_correct(test_obj[key])
            # A pandas index can be numeric, and a dict key can be numeric, but a json key must be a string  # noqa: E501 # FIXME CoP
            new_dict[str(key)] = recursively_convert_to_json_serializable(test_obj[key])

        return new_dict

    elif isinstance(test_obj, (list, tuple, set)):
        new_list = []
        for val in test_obj:
            new_list.append(recursively_convert_to_json_serializable(val))

        return new_list

    elif isinstance(test_obj, (np.ndarray, pd.Index)):
        # test_obj[key] = test_obj[key].tolist()
        # If we have an array or index, convert it first to a list--causing coercion to float--and then round  # noqa: E501 # FIXME CoP
        # to the number of digits for which the string representation will equal the float representation  # noqa: E501 # FIXME CoP
        return [_recursively_convert_to_json_serializable(x) for x in test_obj.tolist()]

    # Note: This clause has to come after checking for np.ndarray or we get:
    #      `ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()`  # noqa: E501 # FIXME CoP
    elif test_obj is None:
        # No problem to encode json
        return test_obj

    elif isinstance(test_obj, (datetime.datetime, datetime.date)):
        return str(test_obj)

    # Use built in base type from numpy, https://docs.scipy.org/doc/numpy-1.13.0/user/basics.types.html
    # https://github.com/numpy/numpy/pull/9505
    elif np.issubdtype(type(test_obj), np.bool_):
        return bool(test_obj)

    elif np.issubdtype(type(test_obj), np.integer) or np.issubdtype(type(test_obj), np.uint):
        return int(test_obj)

    elif np.issubdtype(type(test_obj), np.floating):
        # Note: Use np.floating to avoid FutureWarning from numpy
        return float(round(test_obj, sys.float_info.dig))

    elif isinstance(test_obj, pd.Series):
        # Converting a series is tricky since the index may not be a string, but all json
        # keys must be strings. So, we use a very ugly serialization strategy
        index_name = test_obj.index.name or "index"
        value_name = test_obj.name or "value"
        return [
            {
                index_name: _recursively_convert_to_json_serializable(idx),
                value_name: _recursively_convert_to_json_serializable(val),
            }
            for idx, val in test_obj.items()
        ]

    elif isinstance(test_obj, pd.DataFrame):
        return _recursively_convert_to_json_serializable(test_obj.to_dict(orient="records"))

    # elif np.issubdtype(type(test_obj), np.complexfloating):
    # Note: Use np.complexfloating to avoid Future Warning from numpy
    # Complex numbers consist of two floating point numbers
    # return complex(
    #     float(round(test_obj.real, sys.float_info.dig)),
    #     float(round(test_obj.imag, sys.float_info.dig)))

    elif isinstance(test_obj, decimal.Decimal):
        return float(test_obj)

    else:
        raise TypeError(  # noqa: TRY003 # FIXME CoP
            f"{test_obj!s} is of type {type(test_obj).__name__} which cannot be serialized."
        )


def ensure_row_condition_is_correct(row_condition_string) -> None:
    """Ensure no quote nor \\\\n are introduced in row_condition string.

    Otherwise it may cause an issue at the reload of the expectation.
    An error is raised at the declaration of the expectations to ensure
    the user is not doing a mistake. He can use double quotes for example.

    Parameters
    ----------
    row_condition_string : str
        the pandas query string
    """
    if "'" in row_condition_string:
        raise InvalidExpectationConfigurationError(  # noqa: TRY003 # FIXME CoP
            f"{row_condition_string} cannot be serialized to json. "
            "Do not introduce simple quotes in configuration."
            "Use double quotes instead."
        )
    if "\n" in row_condition_string:
        raise InvalidExpectationConfigurationError(  # noqa: TRY003 # FIXME CoP
            f"{row_condition_string!r} cannot be serialized to json. Do not introduce \\n in configuration."  # noqa: E501 # FIXME CoP
        )



from __future__ import annotations

import abc
import logging
import os
import pathlib
import shutil
import warnings
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from ruamel.yaml import YAML

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.constants import (
    CURRENT_GX_CONFIG_VERSION,
    MINIMUM_SUPPORTED_CONFIG_VERSION,
)
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.data_context.templates import (
    CONFIG_VARIABLES_TEMPLATE,
    PROJECT_TEMPLATE_USAGE_STATISTICS_ENABLED,
)
from great_expectations.data_context.types.base import (
    DataContextConfigDefaults,
)
from great_expectations.data_context.util import file_relative_path

logger = logging.getLogger(__name__)
yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr


class SerializableDataContext(AbstractDataContext):
    UNCOMMITTED_DIRECTORIES: ClassVar[list[str]] = ["data_docs", "validations"]
    GX_UNCOMMITTED_DIR: ClassVar[str] = "uncommitted"
    GITIGNORE: ClassVar[str] = ".gitignore"
    GX_CONFIG_VARIABLES: ClassVar[str] = "config_variables.yml"
    BASE_DIRECTORIES: ClassVar[list[str]] = [
        DataContextConfigDefaults.CHECKPOINTS_BASE_DIRECTORY.value,
        DataContextConfigDefaults.EXPECTATIONS_BASE_DIRECTORY.value,
        DataContextConfigDefaults.PLUGINS_BASE_DIRECTORY.value,
        DataContextConfigDefaults.VALIDATION_DEFINITIONS_BASE_DIRECTORY.value,
        GX_UNCOMMITTED_DIR,
    ]
    GX_DIR: ClassVar[str] = "gx"
    _LEGACY_GX_DIR: ClassVar[str] = "great_expectations"
    GX_YML: ClassVar[str] = "great_expectations.yml"
    GX_EDIT_NOTEBOOK_DIR: ClassVar[str] = GX_UNCOMMITTED_DIR

    def __init__(
        self,
        context_root_dir: PathStr,
        runtime_environment: Optional[dict] = None,
        user_agent_str: Optional[str] = None,
    ) -> None:
        if isinstance(context_root_dir, pathlib.Path):
            # TODO: (kilo59) 122022 should be saving and passing around `pathlib.Path` not str
            context_root_dir = str(context_root_dir)
        self._context_root_directory = context_root_dir
        super().__init__(
            runtime_environment=runtime_environment,
            user_agent_str=user_agent_str,
        )

    def _init_datasource_store(self):  # type: ignore[explicit-override] # FIXME
        raise NotImplementedError  # Required by parent ABC but this class is never instantiated

    def _init_variables(self):  # type: ignore[explicit-override] # FIXME
        raise NotImplementedError  # Required by parent ABC but this class is never instantiated

    @property
    @override
    def root_directory(self) -> str:
        """The root directory for configuration objects in the data context; the location in which
        ``great_expectations.yml`` is located.
        """
        return self._context_root_directory

    @abc.abstractmethod
    @override
    def _save_project_config(self) -> None:
        """
        See parent 'AbstractDataContext._save_project_config()` for more information.
        Explicitly override base class implementation to retain legacy behavior.
        """
        raise NotImplementedError

    @classmethod
    def _resolve_context_root_dir_and_project_root_dir(
        cls, context_root_dir: PathStr | None, project_root_dir: PathStr | None
    ) -> PathStr | None:
        if project_root_dir and context_root_dir:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                "'project_root_dir' and 'context_root_dir' are conflicting args; please only provide one"  # noqa: E501 # FIXME CoP
            )

        if project_root_dir:
            project_root_dir = pathlib.Path(project_root_dir).absolute()
            context_root_dir = pathlib.Path(project_root_dir) / cls.GX_DIR
        elif context_root_dir:
            context_root_dir = pathlib.Path(context_root_dir).absolute()

        return context_root_dir

    @classmethod
    def _create(
        cls,
        project_root_dir: Optional[PathStr] = None,
        runtime_environment: Optional[dict] = None,
    ) -> SerializableDataContext:
        """
        Build a new gx directory and DataContext object in the provided project_root_dir.

        `create` will create a new "gx" directory in the provided folder, provided one does not
        already exist. Then, it will initialize a new DataContext in that folder and write the resulting config.

        Args:
            project_root_dir: path to the root directory in which to create a new gx directory
            usage_statistics_enabled: boolean directive specifying whether or not to gather usage statistics
            runtime_environment: a dictionary of config variables that override both those set in
                config_variables.yml and the environment

        Returns:
            DataContext
        """  # noqa: E501 # FIXME CoP
        gx_dir = cls._scaffold(
            project_root_dir=project_root_dir,
        )
        return cls(context_root_dir=gx_dir, runtime_environment=runtime_environment)

    @classmethod
    def _scaffold(
        cls,
        project_root_dir: Optional[PathStr] = None,
        context_root_dir_name: Optional[str] = None,
    ) -> pathlib.Path:
        if not project_root_dir:
            project_root_dir = pathlib.Path.cwd()
        else:
            project_root_dir = pathlib.Path(project_root_dir)

        if context_root_dir_name is None:
            context_root_dir_name = cls.GX_DIR

        gx_dir = project_root_dir / context_root_dir_name
        gx_dir.mkdir(parents=True, exist_ok=True)
        cls._scaffold_directories(gx_dir)

        if pathlib.Path.is_file(gx_dir.joinpath(cls.GX_YML)):
            message = f"""Warning. An existing `{cls.GX_YML}` was found here: {gx_dir}.
    - No action was taken."""
            warnings.warn(message)
        else:
            cls._write_project_template_to_disk(gx_dir)

        uncommitted_dir = gx_dir / cls.GX_UNCOMMITTED_DIR
        if pathlib.Path.is_file(uncommitted_dir.joinpath(cls.GX_CONFIG_VARIABLES)):
            message = f"""Warning. An existing `config_variables.yml` was found here:
            {uncommitted_dir}. - No action was taken."""
            warnings.warn(message)
        else:
            cls._write_config_variables_template_to_disk(uncommitted_dir)

        return gx_dir

    @classmethod
    def all_uncommitted_directories_exist(cls, gx_dir: PathStr) -> bool:
        """Check if all uncommitted directories exist."""
        gx_dir = pathlib.Path(gx_dir)
        uncommitted_dir = gx_dir / cls.GX_UNCOMMITTED_DIR
        for directory in cls.UNCOMMITTED_DIRECTORIES:
            if not pathlib.Path.is_dir(uncommitted_dir.joinpath(directory)):
                return False

        return True

    @classmethod
    def config_variables_yml_exist(cls, gx_dir: PathStr) -> bool:
        """Check if all config_variables.yml exists."""
        gx_dir = pathlib.Path(gx_dir)
        path_to_yml = gx_dir / cls.GX_YML

        # TODO this is so brittle and gross
        with path_to_yml.open() as f:
            config = yaml.load(f)
        config_var_path = config.get("config_variables_file_path")
        if not config_var_path:
            return False
        config_var_path = pathlib.Path(config_var_path)
        config_var_path = gx_dir / config_var_path
        return config_var_path.is_file()

    @classmethod
    def _write_config_variables_template_to_disk(cls, uncommitted_dir: PathStr) -> None:
        uncommitted_dir = pathlib.Path(uncommitted_dir)

        uncommitted_dir.mkdir(exist_ok=True)
        config_var_file = uncommitted_dir / cls.GX_CONFIG_VARIABLES
        with config_var_file.open("w") as template:
            template.write(CONFIG_VARIABLES_TEMPLATE)

    @classmethod
    def _write_project_template_to_disk(cls, gx_dir: PathStr) -> None:
        gx_dir = pathlib.Path(gx_dir)
        file_path = gx_dir / cls.GX_YML
        with file_path.open("w") as template:
            template.write(PROJECT_TEMPLATE_USAGE_STATISTICS_ENABLED)

    @classmethod
    def _scaffold_directories(cls, base_dir: pathlib.Path) -> None:
        """Safely create GE directories for a new project."""
        base_dir.mkdir(exist_ok=True)

        try:
            cls._scaffold_gitignore(base_dir)
        except Exception as e:
            raise gx_exceptions.GitIgnoreScaffoldingError(  # noqa: TRY003 # FIXME CoP
                f"Could not create .gitignore in {base_dir} because of an error: {e}"
            )

        for directory in cls.BASE_DIRECTORIES:
            if directory == "plugins":
                plugins_dir = base_dir / directory
                plugins_dir.mkdir(exist_ok=True)

                custom_data_docs = plugins_dir / "custom_data_docs"
                custom_data_docs.mkdir(exist_ok=True)

                views = custom_data_docs / "views"
                views.mkdir(exist_ok=True)

                renderers = custom_data_docs / "renderers"
                renderers.mkdir(exist_ok=True)

                styles = custom_data_docs / "styles"
                styles.mkdir(exist_ok=True)

                cls._scaffold_custom_data_docs(plugins_dir)
            else:
                non_plugin_dir = base_dir / directory
                non_plugin_dir.mkdir(exist_ok=True)

        uncommitted_dir = base_dir / cls.GX_UNCOMMITTED_DIR

        for new_directory in cls.UNCOMMITTED_DIRECTORIES:
            new_directory_path = uncommitted_dir / new_directory
            new_directory_path.mkdir(exist_ok=True)

    @classmethod
    def _scaffold_gitignore(cls, base_dir: PathStr) -> None:
        """Make sure .gitignore exists and contains uncommitted/"""
        gitignore = pathlib.Path(base_dir) / cls.GITIGNORE

        uncommitted_dir = f"{cls.GX_UNCOMMITTED_DIR}/"
        if gitignore.is_file():
            contents = gitignore.read_text()
            if uncommitted_dir in contents:
                return

        with gitignore.open("a") as f:
            f.write(f"\n{uncommitted_dir}")

    @classmethod
    def _scaffold_custom_data_docs(cls, plugins_dir: pathlib.Path) -> None:
        """Copy custom data docs templates"""
        styles_template = file_relative_path(
            __file__,
            "../../render/view/static/styles/data_docs_custom_styles_template.css",
        )
        styles_destination_path = (
            plugins_dir / "custom_data_docs" / "styles" / "data_docs_custom_styles.css"
        )
        shutil.copyfile(styles_template, styles_destination_path)

    @classmethod
    def find_context_root_dir(cls) -> str:
        result = None
        yml_path = None
        gx_home_environment = os.getenv("GX_HOME")
        if gx_home_environment:
            gx_home_environment = os.path.expanduser(  # noqa: PTH111 # FIXME CoP
                gx_home_environment
            )
            if os.path.isdir(  # noqa: PTH112 # FIXME CoP
                gx_home_environment
            ) and os.path.isfile(  # noqa: PTH113 # FIXME CoP
                os.path.join(gx_home_environment, cls.GX_YML)  # noqa: PTH118 # FIXME CoP
            ):
                result = gx_home_environment
        else:
            yml_path = cls._find_context_yml_file()
            if yml_path:
                result = os.path.dirname(yml_path)  # noqa: PTH120 # FIXME CoP

        if result is None:
            raise gx_exceptions.ConfigNotFoundError()

        logger.debug(f"Using project config: {yml_path}")
        return result

    @classmethod
    def get_ge_config_version(cls, context_root_dir: Optional[PathStr] = None) -> Optional[float]:
        yml_path = cls._find_context_yml_file(search_start_dir=context_root_dir)
        if yml_path is None:
            return None

        with open(yml_path) as f:
            config_commented_map_from_yaml = yaml.load(f)

        config_version = config_commented_map_from_yaml.get("config_version")
        return float(config_version) if config_version else None

    @classmethod
    def set_ge_config_version(
        cls,
        config_version: Union[int, float],  # noqa: PYI041 # FIXME CoP
        context_root_dir: Optional[str] = None,
        validate_config_version: bool = True,
    ) -> bool:
        if not isinstance(config_version, (int, float)):
            raise gx_exceptions.UnsupportedConfigVersionError(  # noqa: TRY003 # FIXME CoP
                "The argument `config_version` must be a number.",
            )

        if validate_config_version:
            if config_version < MINIMUM_SUPPORTED_CONFIG_VERSION:
                raise gx_exceptions.UnsupportedConfigVersionError(  # noqa: TRY003 # FIXME CoP
                    f"""Invalid config version ({config_version})\n
                                                                  The version number must be at least {MINIMUM_SUPPORTED_CONFIG_VERSION}"""  # noqa: E501 # FIXME CoP
                )
            elif config_version > CURRENT_GX_CONFIG_VERSION:
                raise gx_exceptions.UnsupportedConfigVersionError(  # noqa: TRY003 # FIXME CoP
                    f"""Invalid config version ({config_version}).\n
                                                                  The maximum valid version is {CURRENT_GX_CONFIG_VERSION}."""  # noqa: E501 # FIXME CoP
                )

        yml_path = cls._find_context_yml_file(search_start_dir=context_root_dir)
        if yml_path is None:
            return False

        with open(yml_path) as f:
            config_commented_map_from_yaml = yaml.load(f)
            config_commented_map_from_yaml["config_version"] = float(config_version)

        with open(yml_path, "w") as f:
            yaml.dump(config_commented_map_from_yaml, f)

        return True

    @classmethod
    def _find_context_yml_file(cls, search_start_dir: Optional[PathStr] = None) -> str | None:
        """Search for the yml file starting here and moving upward."""
        if search_start_dir is None:
            search_start_dir = pathlib.Path.cwd()
        else:
            search_start_dir = pathlib.Path(search_start_dir)

        # Ensure backwards compatibility if user is using "great_expectations/" over "gx/"
        # Starting v0.17.13, "gx/" will be the default
        return cls._search_gx_dir_for_context_yml(
            search_start_dir=search_start_dir, gx_dir=cls.GX_DIR
        ) or cls._search_gx_dir_for_context_yml(
            search_start_dir=search_start_dir, gx_dir=cls._LEGACY_GX_DIR
        )

    @classmethod
    def _search_gx_dir_for_context_yml(
        cls, search_start_dir: pathlib.Path, gx_dir: str
    ) -> Optional[str]:
        yml_path: str | None = None

        for i in range(4):
            logger.debug(f"Searching for config file {search_start_dir} ({i} layer deep)")

            potential_ge_dir = search_start_dir / gx_dir

            if potential_ge_dir.is_dir():
                potential_yml = potential_ge_dir / cls.GX_YML
                if potential_yml.is_file():
                    yml_path = str(potential_yml)
                    logger.debug(f"Found config file at {yml_path}")
                    break

            # move up one directory
            search_start_dir = search_start_dir.parent

        return yml_path

    @classmethod
    def does_config_exist_on_disk(cls, context_root_dir: PathStr) -> bool:
        """Return True if the great_expectations.yml exists on disk."""
        context_root_dir = pathlib.Path(context_root_dir)
        config = context_root_dir / cls.GX_YML
        return config.is_file()

    @classmethod
    def is_project_initialized(cls, ge_dir: PathStr) -> bool:
        """
        Return True if the project is initialized.

        To be considered initialized, all of the following must be true:
        - the project must be scaffolded (see cls.is_project_scaffolded)
        - the project has at least one datasource
        - the project has at least one suite
        """
        return (
            cls.is_project_scaffolded(ge_dir)
            and cls._does_context_have_at_least_one_datasource(ge_dir)
            and cls._does_context_have_at_least_one_suite(ge_dir)
        )

    @classmethod
    def is_project_scaffolded(cls, ge_dir: PathStr) -> bool:
        """
        Return True if the project is scaffolded (required filesystem changes have occurred).

        To be considered scaffolded, all of the following must be true:
        - all project directories exist (including uncommitted directories)
        - a valid great_expectations.yml is on disk
        - a config_variables.yml is on disk
        """
        return (
            cls.does_config_exist_on_disk(ge_dir)
            and cls.all_uncommitted_directories_exist(ge_dir)
            and cls.config_variables_yml_exist(ge_dir)
        )

    @classmethod
    def _does_project_have_a_datasource_in_config_file(cls, ge_dir: PathStr) -> bool:
        if not cls.does_config_exist_on_disk(ge_dir):
            return False
        return cls._does_context_have_at_least_one_datasource(ge_dir)

    @classmethod
    def _does_context_have_at_least_one_datasource(cls, ge_dir: PathStr) -> bool:
        context = cls._attempt_context_instantiation(ge_dir)
        if not context:
            return False
        return len(context.list_datasources()) >= 1

    @classmethod
    def _does_context_have_at_least_one_suite(cls, ge_dir: PathStr) -> bool:
        context = cls._attempt_context_instantiation(ge_dir)
        if not context:
            return False
        return bool(context.suites.all())

    @classmethod
    def _attempt_context_instantiation(cls, ge_dir: PathStr) -> Optional[SerializableDataContext]:
        try:
            context = cls(context_root_dir=ge_dir)
            return context
        except (
            gx_exceptions.DataContextError,
            gx_exceptions.InvalidDataContextConfigError,
        ) as e:
            logger.debug(e)
        return None



from typing import Union

from great_expectations.compatibility.pydantic import (
    BaseModel,
)


class _IdentifierBundle(BaseModel):
    name: str
    id: Union[str, None]


class _EncodedValidationData(BaseModel):
    datasource: _IdentifierBundle
    asset: _IdentifierBundle
    batch_definition: _IdentifierBundle



from __future__ import annotations

import enum
import uuid
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from great_expectations.compatibility.pydantic import BaseModel, Field
from great_expectations.compatibility.typing_extensions import override

if TYPE_CHECKING:
    MappingIntStrAny = Mapping[Union[int, str], Any]
    AbstractSetIntStr = AbstractSet[Union[int, str]]


class MetricTypesMeta(enum.EnumMeta):
    """Metaclass definition for MetricTypes that allows for membership checking."""

    def __contains__(cls, item):  # type: ignore[explicit-override] # FIXME
        return item in cls.__members__.values()


class MetricTypes(str, enum.Enum, metaclass=MetricTypesMeta):
    """Represents Metric types in OSS that are used for ColumnDescriptiveMetrics and MetricRepository.

    More Metric types will be added in the future.
    """  # noqa: E501 # FIXME CoP

    # Table metrics
    TABLE_COLUMNS = "table.columns"
    TABLE_ROW_COUNT = "table.row_count"
    TABLE_COLUMN_TYPES = "table.column_types"

    # Column metrics
    COLUMN_MIN = "column.min"
    COLUMN_MAX = "column.max"
    COLUMN_MEDIAN = "column.median"
    COLUMN_MEAN = "column.mean"
    COLUMN_NULL_COUNT = "column_values.null.count"


class MetricRepositoryBaseModel(BaseModel):
    """Base class for all MetricRepository related models."""

    class Config:
        extra = "forbid"


class MetricException(MetricRepositoryBaseModel):
    type: str = Field(description="Exception type if an exception is thrown")
    message: str = Field(description="Exception message if an exception is thrown")


_ValueType = TypeVar("_ValueType")


class Metric(MetricRepositoryBaseModel, Generic[_ValueType]):
    """Abstract computed metric. Domain, value and parameters are metric dependent.

    Note: This implementation does not currently take into account
    other domain modifiers, e.g. row_condition, condition_parser, ignore_row_if
    """

    def __new__(cls, *args, **kwargs):
        if cls is Metric:
            raise NotImplementedError("Metric is an abstract class.")
        instance = super().__new__(cls)
        return instance

    batch_id: str = Field(description="Batch id")
    metric_name: str = Field(description="Metric name")
    value: _ValueType = Field(description="Metric value")
    exception: Optional[MetricException] = Field(
        description="Exception info if thrown", default=None
    )

    @classmethod
    def update_forward_refs(cls):
        from great_expectations.datasource.fluent.interfaces import Batch

        super().update_forward_refs(
            Batch=Batch,
        )

    @property
    def value_type(self) -> str:
        type_ = self.__orig_class__.__args__[0]  # type: ignore[attr-defined] # __orig_class__ is used to get the generic type
        string_rep = str(type_)
        if string_rep.startswith("<class"):
            return type_.__name__
        else:
            return string_rep

    @property
    def metric_type(self) -> str:
        return self.__class__.__name__

    @classmethod
    def _get_properties(cls):
        """in pydandic v2 we can use computed_field.
        https://docs.pydantic.dev/latest/usage/computed_fields/"""
        properties = [prop for prop in cls.__dict__ if isinstance(cls.__dict__[prop], property)]
        return properties

    @override
    def dict(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        """Override the dict function to include @property fields, in pydandic v2 we can use computed_field.
        https://docs.pydantic.dev/latest/usage/computed_fields/
        """  # noqa: E501 # FIXME CoP
        attribs = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
        props = self._get_properties()

        # Include and exclude properties
        if include:
            props = [prop for prop in props if prop in include]
        if exclude:
            props = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})
        return attribs


# Metric domain types


class TableMetric(Metric, Generic[_ValueType]):
    @override
    @property
    def value_type(self) -> str:
        type_ = self.__orig_class__.__args__[0]  # type: ignore[attr-defined] # __orig_class__ is used to get the generic type
        string_rep = str(type_)
        if string_rep.startswith("<class"):
            return type_.__name__
        else:
            return string_rep

    @override
    @property
    def metric_type(self) -> str:
        return self.__class__.__name__


class ColumnMetric(Metric, Generic[_ValueType]):
    column: str = Field(description="Column name")

    @override
    @property
    def value_type(self) -> str:
        type_ = self.__orig_class__.__args__[0]  # type: ignore[attr-defined] # __orig_class__ is used to get the generic type
        string_rep = str(type_)
        if string_rep.startswith("<class"):
            return type_.__name__
        else:
            return string_rep

    @override
    @property
    def metric_type(self) -> str:
        return self.__class__.__name__


# TODO: Add ColumnPairMetric, MultiColumnMetric


# Metrics with parameters (aka metric_value_kwargs)
# This is where the concrete metric types are defined that
# bring together a domain type, value type and any parameters (aka metric_value_kwargs)
# If a metric has parameters, it should be defined here. If it doesn't, you can use the generic types above, for  # noqa: E501 # FIXME CoP
# example, ColumnMetric[float] or TableMetric[list[str]].

# TODO: Add metrics here for all Column Descriptive Metrics
#  ColumnQuantileValuesMetric is an example of a metric that has parameters


class ColumnQuantileValuesMetric(ColumnMetric[List[float]]):
    quantiles: List[float] = Field(description="Quantiles to compute")
    allow_relative_error: Union[float, str] = Field(
        description="Relative error interpolation type (pandas) or limit (e.g. spark) depending on data source"  # noqa: E501 # FIXME CoP
    )

    @property
    @override
    def value_type(self) -> str:
        return "list[float]"

    @property
    @override
    def metric_type(self) -> str:
        return self.__class__.__name__


class MetricRun(MetricRepositoryBaseModel):
    """Collection of Metric objects produced during the same execution run."""

    data_asset_id: Union[uuid.UUID, None] = Field(description="Data asset id", default=None)
    metrics: Sequence[Metric]



import pytest

from great_expectations.util import (
    get_clickhouse_sqlalchemy_potential_type,
    is_library_loadable,
)
from tests.test_utils import get_awsathena_connection_url


@pytest.mark.athena
def test_get_awsathena_connection_url(monkeypatch):
    monkeypatch.setenv("ATHENA_STAGING_S3", "s3://test-staging/")
    monkeypatch.setenv("ATHENA_DB_NAME", "test_db_name")
    monkeypatch.setenv("ATHENA_TEN_TRIPS_DB_NAME", "test_ten_trips_db_name")

    assert (
        get_awsathena_connection_url()
        == "awsathena+rest://@athena.us-east-1.amazonaws.com/test_db_name?s3_staging_dir=s3://test-staging/"
    )

    assert (
        get_awsathena_connection_url(db_name_env_var="ATHENA_TEN_TRIPS_DB_NAME")
        == "awsathena+rest://@athena.us-east-1.amazonaws.com/test_ten_trips_db_name?s3_staging_dir=s3://test-staging/"
    )


@pytest.mark.clickhouse
@pytest.mark.skipif(
    not is_library_loadable(library_name="clickhouse_sqlalchemy"),
    reason="clickhouse_sqlalchemy is not installed",
)
def test_get_clickhouse_sqlalchemy_potential_type():
    import clickhouse_sqlalchemy
    from clickhouse_sqlalchemy import types

    input_output = (
        ("Nullable(String)", types.String),
        ("Int8", types.Int8),
        ("Map(String, String)", types.Map),
    )
    for pair in input_output:
        assert (
            get_clickhouse_sqlalchemy_potential_type(clickhouse_sqlalchemy.drivers.base, pair[0])
            == pair[1]
        )



from __future__ import annotations

import json
from typing import NamedTuple, Optional, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.domain import Domain
from great_expectations.core.id_dict import IDDict, IDDictID
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.experimental.metric_repository.metrics import MetricTypes
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP


class MetricConfigurationID(NamedTuple):
    metric_name: str
    metric_domain_kwargs_id: IDDictID
    metric_value_kwargs_id: IDDictID


class MetricConfiguration:
    """An interface for configuring Metrics.

    MetricConfiguration allows the configuration of domain information, dependencies and additional metric-specific
    configurations.  Metrics are computed attributes of data, and are derived from one or more Batches that can then
    be used to evaluate Expectations or to summarize the result of the Validation.

    Args:
        metric_name (str or MetricTypes enum): name of the Metric defined by the current MetricConfiguration.
        metric_domain_kwargs (dict): provides information on where the Metric can be calculated. For instance, a
            MapCondition metric can include the name of the column that the Metric is going to be run on.
        metric_value_kwargs (optional[dict]): Optional kwargs that define values specific to each Metric.  For instance,
            a Metric that partitions a column can define the method of partitioning (`uniform` bins) and the number of
            bins (`n_bins`) as `metric_value_kwargs`.
    """  # noqa: E501 # FIXME CoP

    def __init__(
        self,
        metric_name: Union[str, MetricTypes],
        metric_domain_kwargs: dict,
        metric_value_kwargs: Optional[dict] = None,
    ) -> None:
        if isinstance(metric_name, MetricTypes):
            metric_name = metric_name.value

        if not isinstance(metric_domain_kwargs, IDDict):
            metric_domain_kwargs = IDDict(metric_domain_kwargs)

        self._metric_domain_kwargs: IDDict = metric_domain_kwargs

        if not isinstance(metric_value_kwargs, IDDict):
            if metric_value_kwargs is None:
                metric_value_kwargs = {}
            metric_value_kwargs = IDDict(metric_value_kwargs)

        self._metric_value_kwargs: IDDict = metric_value_kwargs

        self._metric_dependencies: IDDict = IDDict({})

        self._id = MetricConfigurationID(
            metric_name,
            metric_domain_kwargs.to_id(),
            metric_value_kwargs.to_id(),
        )

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return json.dumps(self.to_json_dict(), indent=2)

    @override
    def __str__(self):
        return self.__repr__()

    @property
    def metric_name(self) -> str:
        return self.id.metric_name

    @property
    def metric_domain_kwargs(self) -> IDDict:
        return self._metric_domain_kwargs

    @property
    def metric_value_kwargs(self) -> IDDict:
        return self._metric_value_kwargs

    @property
    def metric_domain_kwargs_id(self) -> IDDictID:
        return self.id.metric_domain_kwargs_id

    @property
    def metric_value_kwargs_id(self) -> IDDictID:
        return self.id.metric_value_kwargs_id

    @property
    def metric_dependencies(self) -> IDDict:
        return self._metric_dependencies

    @metric_dependencies.setter
    def metric_dependencies(self, metric_dependencies: dict) -> None:
        if metric_dependencies is None:
            metric_dependencies = IDDict({})

        if not isinstance(metric_dependencies, IDDict):
            metric_dependencies = IDDict(metric_dependencies)

        self._metric_dependencies = metric_dependencies

    def get_domain(self) -> Domain:
        """Return "Domain" object, constructed from this "MetricConfiguration" object."""
        domain_type: MetricDomainTypes = self.get_domain_type()

        if domain_type == MetricDomainTypes.TABLE:
            other_table_name: Optional[str] = self._metric_domain_kwargs.get("table")
            if other_table_name:
                return Domain(
                    domain_type=domain_type,
                    domain_kwargs={
                        "table": other_table_name,
                    },
                )

            return Domain(
                domain_type=domain_type,
            )

        if domain_type == MetricDomainTypes.COLUMN:
            return Domain(
                domain_type=domain_type,
                domain_kwargs={
                    "column": self._metric_domain_kwargs["column"],
                },
            )

        if domain_type == MetricDomainTypes.COLUMN_PAIR:
            return Domain(
                domain_type=domain_type,
                domain_kwargs={
                    "column_A": self._metric_domain_kwargs["column_A"],
                    "column_B": self._metric_domain_kwargs["column_B"],
                },
            )

        if domain_type == MetricDomainTypes.MULTICOLUMN:
            return Domain(
                domain_type=domain_type,
                domain_kwargs={
                    "column_list": self._metric_domain_kwargs["column_list"],
                },
            )

        raise ValueError(f"""Domain type "{domain_type}" is not recognized.""")  # noqa: TRY003 # FIXME CoP

    def get_domain_type(self) -> MetricDomainTypes:
        """Return "domain_type" of this "MetricConfiguration" object."""
        if "column" in self._metric_domain_kwargs:
            return MetricDomainTypes.COLUMN

        if "column_A" in self._metric_domain_kwargs and "column_B" in self._metric_domain_kwargs:
            return MetricDomainTypes.COLUMN_PAIR

        if "column_list" in self._metric_domain_kwargs:
            return MetricDomainTypes.MULTICOLUMN

        # TODO: <Alex>Determining "domain_type" of "MetricConfiguration" using heuristics defaults to "TABLE".</Alex>  # noqa: E501 # FIXME CoP
        return MetricDomainTypes.TABLE

    @property
    def id(self) -> MetricConfigurationID:
        return self._id

    def to_json_dict(self) -> dict:
        """Returns a JSON-serializable dict representation of this MetricConfiguration.

        Returns:
            A JSON-serializable dict representation of this MetricConfiguration.

        """
        json_dict: dict = convert_to_json_serializable(
            data={
                "metric_name": self.metric_name,
                "metric_domain_kwargs": self.metric_domain_kwargs,
                "metric_domain_kwargs_id": self.metric_domain_kwargs_id,
                "metric_value_kwargs": self.metric_value_kwargs,
                "metric_value_kwargs_id": self.metric_value_kwargs_id,
                "id": self.id,
            }
        )
        return json_dict



from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

multiple_backend = []

connecting_to_your_data: list[IntegrationTestFixture] = []

deployment_patterns: list[IntegrationTestFixture] = []

creating_custom_expectations = [
    IntegrationTestFixture(
        name="expect_column_values_to_equal_three",
        user_flow_script="docs/docusaurus/docs/snippets/expect_column_values_to_equal_three.py",
        backend_dependencies=[
            BackendDependencies.SPARK,
            BackendDependencies.POSTGRESQL,
        ],
    ),
    IntegrationTestFixture(
        name="expect_column_pair_values_to_have_a_difference_of_three",
        user_flow_script="docs/docusaurus/docs/oss/guides/expectations/creating_custom_expectations/expect_column_pair_values_to_have_a_difference_of_three.py",
        backend_dependencies=[
            BackendDependencies.SPARK,
            BackendDependencies.POSTGRESQL,
        ],
    ),
]

fluent_datasources = [
    IntegrationTestFixture(
        name="how_to_connect_to_data_on_s3_using_spark",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/filesystem/how_to_connect_to_data_on_s3_using_spark.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.AWS, BackendDependencies.SPARK],
    ),
    IntegrationTestFixture(
        name="how_to_connect_to_data_on_gcs_using_spark",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/filesystem/how_to_connect_to_data_on_gcs_using_spark.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.GCS, BackendDependencies.SPARK],
    ),
    IntegrationTestFixture(
        name="how_to_connect_to_data_on_azure_blob_storage_using_spark",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/filesystem/how_to_connect_to_data_on_azure_blob_storage_using_spark.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.AZURE, BackendDependencies.SPARK],
    ),
]

multiple_backend += connecting_to_your_data
multiple_backend += deployment_patterns
multiple_backend += creating_custom_expectations
multiple_backend += fluent_datasources



import pytest

from great_expectations.compatibility.not_imported import NotImported

# module level markers
pytestmark = pytest.mark.unit


@pytest.fixture
def not_imported() -> NotImported:
    return NotImported("Please install foo")


def test_get_attr_on_not_installed(not_imported):
    with pytest.raises(ModuleNotFoundError, match=str(not_imported)):
        _ = not_imported.thing


def test_set_attr_on_not_installed(not_imported):
    with pytest.raises(ModuleNotFoundError, match=str(not_imported)):
        not_imported.thing = 5


def test_calling_not_installed(not_imported):
    with pytest.raises(ModuleNotFoundError, match=str(not_imported)):
        not_imported()



import pathlib
from typing import Mapping, Optional

import pandas as pd
import pytest

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)
from tests.integration.test_utils.data_source_config.sql import SQLBatchTestSetup


class SqliteDatasourceTestConfig(DataSourceTestConfig):
    @property
    @override
    def label(self) -> str:
        return "sqlite"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.sqlite

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        tmp_path = request.getfixturevalue("tmp_path")
        assert isinstance(tmp_path, pathlib.Path)

        return SqliteBatchTestSetup(
            data=data,
            config=self,
            base_dir=tmp_path,
            extra_data=extra_data,
            table_name=self.table_name,
            context=context,
            engine_manager=engine_manager,
        )


class SqliteBatchTestSetup(SQLBatchTestSetup[SqliteDatasourceTestConfig]):
    def __init__(
        self,
        data: pd.DataFrame,
        config: SqliteDatasourceTestConfig,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        base_dir: pathlib.Path,
        table_name: Optional[str] = None,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> None:
        self._base_dir = base_dir
        super().__init__(
            config=config,
            data=data,
            extra_data=extra_data,
            table_name=table_name,
            engine_manager=engine_manager,
            context=context,
        )

    @property
    @override
    def connection_string(self) -> str:
        return f"sqlite:///{self.db_file_path}"

    @property
    @override
    def use_schema(self) -> bool:
        return False

    @property
    def db_file_path(self) -> pathlib.Path:
        return self._base_dir / "database.db"

    @override
    def make_asset(self) -> TableAsset:
        return self.context.data_sources.add_sqlite(
            name=self._random_resource_name(),
            connection_string=self.connection_string,
        ).add_table_asset(name=self._random_resource_name(), table_name=self.table_name)



from typing import Sequence

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

COL_NAME = "col_name"


DATA = pd.DataFrame({COL_NAME: ["aa", "ab", "ac", None]})

REGULAR_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


class TestNormalSql:
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["bc"]
                ),
                id="one_pattern",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["bc", "%de%"]
                ),
                id="multiple_patterns",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=REGULAR_DATA_SOURCES, data=DATA)
    def test_success(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePatternList,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert result.success

    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["%a%"]
                ),
                id="one_pattern",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["%a%", "not_this"]
                ),
                id="multiple_patterns",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=REGULAR_DATA_SOURCES, data=DATA)
    def test_failure(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePatternList,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert not result.success


class TestMSSQL:
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["bc"]
                ),
                id="one_pattern",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["bc", "%de%"]
                ),
                id="multiple_patterns",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=[MSSQLDatasourceTestConfig()], data=DATA
    )
    def test_success(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePatternList,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert result.success

    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["%a[b]%"]
                ),
                id="one_pattern",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePatternList(
                    column=COL_NAME, like_pattern_list=["%[a]%", "not_this"]
                ),
                id="multiple_patterns",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=[MSSQLDatasourceTestConfig()], data=DATA
    )
    def test_failure(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePatternList,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert not result.success



"""This file is meant for integration tests related to datasource CRUD."""

from __future__ import annotations

import random
import string
import urllib.parse

import pytest

import great_expectations as gx
from great_expectations.data_context import CloudDataContext
from great_expectations.data_context.types.base import GXCloudConfig
from great_expectations.datasource.fluent import PandasDatasource

# module level markers
pytestmark = pytest.mark.cloud


def test_cloud_context_add_datasource_with_individual_fds_args_raises_error(
    empty_base_data_context_in_cloud_mode: CloudDataContext,
):
    context = empty_base_data_context_in_cloud_mode

    with pytest.raises(TypeError):
        context.add_datasource(name="my_pandas_fds", type="pandas", assets=[])


def test_cloud_context_add_datasource_with_fds(
    cloud_api_fake,
    empty_cloud_data_context: CloudDataContext,
    ge_cloud_config: GXCloudConfig,
):
    context = empty_cloud_data_context
    name = "my_pandas_ds"

    post_url = urllib.parse.urljoin(
        ge_cloud_config.base_url,
        f"api/v1/organizations/{ge_cloud_config.organization_id}/datasources",
    )

    fds = PandasDatasource(name=name)
    _ = context.add_datasource(datasource=fds)

    assert cloud_api_fake.assert_call_count(url=post_url, count=2)


def test_cloud_context_datasource_crud_e2e() -> None:
    context = gx.get_context(cloud_mode=True)
    datasource_name = f"OSSTestDatasource_{''.join(random.choice(string.ascii_letters + string.digits) for _ in range(8))}"  # noqa: E501 # FIXME CoP

    context.data_sources.add_pandas(name=datasource_name)

    saved_datasource = context.data_sources.get(datasource_name)
    assert saved_datasource is not None and saved_datasource.name == datasource_name

    context.delete_datasource(datasource_name)

    # Make another call to the backend to confirm deletion
    with pytest.raises(KeyError):
        context.data_sources.get(datasource_name)



import pandas as pd
import pytest

from great_expectations import get_context
from great_expectations.compatibility.aws import REDSHIFT_TYPES
from great_expectations.expectations import (
    ExpectColumnValuesToBeOfType,
)
from tests.integration.test_utils.data_source_config import RedshiftDatasourceTestConfig
from tests.integration.test_utils.data_source_config.redshift import RedshiftBatchTestSetup


class TestRedshiftDataTypes:
    """This set of tests ensures that we can run expectations against every data
    type supported by Redshift.

    """

    COLUMN = "col_a"

    @pytest.mark.redshift
    def test_geometry(self):
        column_type = REDSHIFT_TYPES.GEOMETRY
        batch_setup = RedshiftBatchTestSetup(
            config=RedshiftDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame(
                {
                    self.COLUMN: [
                        "0103000020E61000000100000005000000000000000000000000000000000000000000000000000000000000000000F03F000000000000F03F000000000000F03F000000000000F03F000000000000000000000000000000000000000000000000"
                    ]
                }
            ),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeOfType(
                    column=self.COLUMN,
                    type_="GEOMETRY",
                )
            )
        assert result.success

    @pytest.mark.redshift
    def test_super(self):
        column_type = REDSHIFT_TYPES.SUPER
        batch_setup = RedshiftBatchTestSetup(
            config=RedshiftDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: ['{ "type": "Point", "coordinates": [1.0, 2.0] }']}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeOfType(
                    column=self.COLUMN,
                    type_="SUPER",
                )
            )
        assert result.success



from __future__ import annotations

import logging
import uuid
from collections import UserDict
from typing import TYPE_CHECKING, Protocol, TypeVar, runtime_checkable

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent import Datasource as FluentDatasource
from great_expectations.datasource.fluent.constants import _IN_MEMORY_DATA_ASSET_TYPE

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )
    from great_expectations.data_context.store.datasource_store import DatasourceStore
    from great_expectations.datasource.fluent.interfaces import DataAsset

T = TypeVar("T", bound=FluentDatasource)

logger = logging.getLogger(__name__)


@runtime_checkable
class SupportsInMemoryDataAssets(Protocol):
    @property
    def assets(self) -> list[DataAsset]: ...

    def add_dataframe_asset(self, **kwargs) -> DataAsset: ...


class DatasourceDict(UserDict):
    """
    An abstraction around the DatasourceStore to enable easy retrieval and storage of Datasource objects
    using dictionary syntactic sugar.

    Example:
    ```
    d = DatasourceDict(...)

    d["my_fds"] = pandas_fds # Underlying DatasourceStore makes a `set()` call
    pandas_fds = d["my_fds"] # Underlying DatasourceStore makes a `get()` call
    ```
    """  # noqa: E501 # FIXME CoP

    def __init__(
        self,
        context: AbstractDataContext,
        datasource_store: DatasourceStore,
    ):
        self._context = context  # If possible, we should avoid passing the context through - once block-style is removed, we can extract this  # noqa: E501 # FIXME CoP
        self._datasource_store = datasource_store
        self._in_memory_data_assets: dict[str, DataAsset] = {}

    @staticmethod
    def _get_in_memory_data_asset_name(datasource_name: str, data_asset_name: str) -> str:
        return f"{datasource_name}-{data_asset_name}"

    @override
    @property
    def data(self) -> dict[str, FluentDatasource]:  # type: ignore[override] # `data` is meant to be a writeable attr (not a read-only property)
        """
        `data` is referenced by the parent `UserDict` and enables the class to fulfill its various dunder methods
        (__setitem__, __getitem__, etc)

        This is generated just-in-time as the contents of the store may have changed.
        """  # noqa: E501 # FIXME CoP
        datasources: dict[str, FluentDatasource] = {}

        configs = self._datasource_store.get_all()
        for config in configs:
            name = config.name
            try:
                datasources[name] = self._init_fluent_datasource(name=name, ds=config)
            except gx_exceptions.DatasourceInitializationError as e:
                logger.warning(f"Cannot initialize datasource {name}: {e}")

        return datasources

    def set_datasource(self, name: str, ds: FluentDatasource) -> FluentDatasource | None:
        config = self._prep_fds_config_for_set(name=name, ds=ds)

        datasource = self._datasource_store.set(key=None, value=config)
        return self._init_fluent_datasource(name=name, ds=datasource)

    @override
    def __setitem__(self, name: str, ds: FluentDatasource) -> None:
        self.set_datasource(name=name, ds=ds)

    def _prep_fds_config_for_set(self, name: str, ds: FluentDatasource) -> FluentDatasource:
        if isinstance(ds, SupportsInMemoryDataAssets):
            for asset in ds.assets:
                if asset.type == _IN_MEMORY_DATA_ASSET_TYPE:
                    in_memory_asset_name: str = DatasourceDict._get_in_memory_data_asset_name(
                        datasource_name=name,
                        data_asset_name=asset.name,
                    )
                    self._in_memory_data_assets[in_memory_asset_name] = asset
        return ds

    def _get_ds_from_store(self, name: str) -> FluentDatasource:
        try:
            return self._datasource_store.retrieve_by_name(name)
        except ValueError:
            raise KeyError(f"Could not find a datasource named '{name}'")  # noqa: TRY003 # FIXME CoP

    @override
    def __delitem__(self, name: str) -> None:
        ds = self._get_ds_from_store(name)
        self._datasource_store.delete(ds)

    @override
    def __getitem__(self, name: str) -> FluentDatasource:
        ds = self._get_ds_from_store(name)

        return self._init_fluent_datasource(name=name, ds=ds)

    def _init_fluent_datasource(self, name: str, ds: FluentDatasource) -> FluentDatasource:
        ds._data_context = self._context
        ds._rebuild_asset_data_connectors()
        if isinstance(ds, SupportsInMemoryDataAssets):
            for asset in ds.assets:
                if asset.type == _IN_MEMORY_DATA_ASSET_TYPE:
                    in_memory_asset_name: str = DatasourceDict._get_in_memory_data_asset_name(
                        datasource_name=name,
                        data_asset_name=asset.name,
                    )
                    self._in_memory_data_assets[in_memory_asset_name] = asset
        return ds


class CacheableDatasourceDict(DatasourceDict):
    """
    Extends the capabilites of the DatasourceDict by placing a caching layer in front of the underlying store.

    Any retrievals will firstly check an in-memory dictionary before requesting from the store. Other CRUD methods will ensure that
    both cache and store are kept in sync.
    """  # noqa: E501 # FIXME CoP

    def __init__(
        self,
        context: AbstractDataContext,
        datasource_store: DatasourceStore,
    ):
        super().__init__(
            context=context,
            datasource_store=datasource_store,
        )
        self._cache: dict[str, FluentDatasource] = {}

    @override
    @property
    def data(self) -> dict[str, FluentDatasource]:  # type: ignore[override] # `data` is meant to be a writeable attr (not a read-only property)
        return self._cache

    @override
    def __contains__(self, name: object) -> bool:
        if name in self.data:
            return True
        try:
            # Resort to store only if not in cache
            _ = self._get_ds_from_store(str(name))
            return True
        except KeyError:
            return False

    @override
    def set_datasource(self, name: str, ds: FluentDatasource) -> FluentDatasource | None:
        self.data[name] = self._add_ids(ds)
        return ds

    def _add_ids(self, ds: FluentDatasource) -> FluentDatasource:
        # File and ephemeral contexts do not use the store, so we need to add IDs here.
        # Note that this is used for both `add` and `update` operations.
        if ds.id is None:
            ds.id = uuid.uuid4()
        for asset in ds.assets:
            if asset.id is None:
                asset.id = uuid.uuid4()
            for batch_definition in asset.batch_definitions:
                if batch_definition.id is None:
                    batch_definition.id = uuid.uuid4()

        return ds

    @override
    def __delitem__(self, name: str) -> None:
        self.data.pop(name, None)

    @override
    def __getitem__(self, name: str) -> FluentDatasource:
        if name in self.data:
            return self.data[name]

        # Upon cache miss, retrieve from store and add to cache
        ds = super().__getitem__(name)
        self.data[name] = ds
        return ds



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional

from great_expectations.compatibility.typing_extensions import override
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.table_metric_provider import (
    TableMetricProvider,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


class TableColumns(TableMetricProvider):
    metric_name = "table.columns"

    @metric_value(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        column_metadata = metrics["table.column_types"]
        return [col["name"] for col in column_metadata]

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        column_metadata = metrics["table.column_types"]
        return [col["name"] for col in column_metadata]

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        column_metadata = metrics["table.column_types"]
        return [col["name"] for col in column_metadata]

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        dependencies: dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )
        table_domain_kwargs: dict = {
            k: v for k, v in metric.metric_domain_kwargs.items() if k != "column"
        }
        metric_value_kwargs = {
            "include_nested": True,
        }
        dependencies["table.column_types"] = MetricConfiguration(
            metric_name="table.column_types",
            metric_domain_kwargs=table_domain_kwargs,
            metric_value_kwargs=metric.metric_value_kwargs or metric_value_kwargs,
        )
        return dependencies



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
from packaging import version

from great_expectations.compatibility import pydantic, pyspark
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001, RUF100 # FIXME CoP
)
from great_expectations.execution_engine.sqlalchemy_dialect import (
    GXSqlDialect,
)
from great_expectations.expectations.core.expect_column_values_to_be_of_type import (
    _get_potential_sqlalchemy_types,
    _native_type_type_map,
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import COLUMN_DESCRIPTION
from great_expectations.expectations.registry import get_metric_kwargs
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import (
        ExecutionEngine,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs
    from great_expectations.validator.validator import ValidationDependencies

logger = logging.getLogger(__name__)

EXPECTATION_SHORT_DESCRIPTION = "Expect a column to contain values from a specified type list."
TYPE_LIST_DESCRIPTION = """
    A list of strings representing the data type that each column should have as entries. \
    Valid types are defined by the current backend implementation and are dynamically loaded.
    """
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.REDSHIFT.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.SCHEMA.value]


class ExpectColumnValuesToBeInTypeList(ColumnMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnValuesToBeInTypeList is a \
    Column Map Expectation \
    for typed-column backends, and also for Pandas Datasources where the column dtype provides an \
    unambiguous constraints (any dtype except 'object').

    For Pandas columns with dtype of 'object' ExpectColumnValuesToBeInTypeList will \
    independently check each row's type.

    Column Map Expectations are one of the most common types of Expectation.
    They are evaluated for a single column and ask a yes/no question for every row in that column.
    Based on the result, they then calculate the percentage of rows that gave a positive answer. If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        type_list (list[str] or None): \
            {TYPE_LIST_DESCRIPTION}
            For example, valid types for Pandas Datasources include any numpy dtype values \
            (such as 'int64') or native python types (such as 'int'), whereas valid types for a \
            SqlAlchemy Datasource include types named by the current driver such as 'INTEGER' \
            in most SQL dialects and 'TEXT' in dialects such as postgresql. Valid types for \
            Spark Datasources include 'StringType', 'BooleanType' and other pyspark-defined type names.

    Other Parameters:
        mostly (None or a float between 0 and 1): \
            Successful if at least mostly fraction of values match the expectation. \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See also:
        [ExpectColumnValuesToBeOfType](https://greatexpectations.io/expectations/expect_column_values_to_be_of_type)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	"12345" 1
            1 	"abcde" 2
            2 	"1b3d5" 3

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnValuesToBeInTypeList(
                    column="test2",
                    type_list=["NUMBER", "STRING"]
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 0,
                    "unexpected_percent": 0.0,
                    "partial_unexpected_list": [],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 0.0,
                    "unexpected_percent_nonmissing": 0.0
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnValuesToBeInTypeList(
                    column="test",
                    type_list=["NUMBER", "DOUBLE"]
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 3,
                    "unexpected_percent": 100.0,
                    "partial_unexpected_list": [
                        "12345",
                        "abcde",
                        "1b3d5"
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 100.0,
                    "unexpected_percent_nonmissing": 100.0
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    type_list: Union[List[str], SuiteParameterDict, None] = pydantic.Field(
        description=TYPE_LIST_DESCRIPTION
    )

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "column_values.in_type_list"
    domain_keys: ClassVar[Tuple[str, ...]] = (
        "column",
        "row_condition",
        "condition_parser",
    )

    success_keys = (
        "type_list",
        "mostly",
    )
    args_keys = (
        "column",
        "type_list",
    )

    class Config:
        title = "Expect column values to be in type list"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnValuesToBeInTypeList]
        ) -> None:
            ColumnMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    @override
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("type_list", RendererValueType.ARRAY),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if params.type_list:
            array_param_name = "type_list"
            param_prefix = "v__"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            values_string: str = cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )

            if params.mostly and params.mostly.value < 1.0:
                renderer_configuration = cls._add_mostly_pct_param(
                    renderer_configuration=renderer_configuration
                )
                template_str = (
                    "value types must belong to this set: "
                    + values_string
                    + ", at least $mostly_pct % of the time."
                )
            else:
                template_str = f"value types must belong to this set: {values_string}."
        else:
            template_str = "value types may be any value, but observed value will be reported"

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @override
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(  # noqa: C901, PLR0912
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs if configuration else {},
            ["column", "type_list", "mostly", "row_condition", "condition_parser"],
        )

        if params["type_list"] is not None:
            for i, v in enumerate(params["type_list"]):
                params[f"v__{i!s}"] = v
            values_string = " ".join([f"$v__{i!s}" for i, v in enumerate(params["type_list"])])

            if params["mostly"] is not None:
                if isinstance(params["mostly"], (int, float)) and params["mostly"] < 1.0:
                    params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
                    # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")  # noqa: E501 # FIXME CoP
                if include_column_name:
                    template_str = (
                        "$column value types must belong to this set: "
                        + values_string
                        + ", at least $mostly_pct % of the time."
                    )
                else:
                    template_str = (
                        "value types must belong to this set: "
                        + values_string
                        + ", at least $mostly_pct % of the time."
                    )
            else:  # noqa: PLR5501 # FIXME CoP
                if include_column_name:
                    template_str = f"$column value types must belong to this set: {values_string}."
                else:
                    template_str = f"value types must belong to this set: {values_string}."
        else:  # noqa: PLR5501 # FIXME CoP
            if include_column_name:
                template_str = (
                    "$column value types may be any value, but observed value will be reported"
                )
            else:
                template_str = "value types may be any value, but observed value will be reported"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    def _validate_pandas(  # noqa: C901, PLR0912 # FIXME CoP
        self,
        actual_column_type,
        expected_types_list,
    ):
        if expected_types_list is None:
            success = True
        else:
            comp_types = []
            for type_ in expected_types_list:
                try:
                    comp_types.append(np.dtype(type_).type)
                    comp_types.append(np.dtype(type_))
                except TypeError:
                    try:
                        pd_type = getattr(pd, type_)
                    except AttributeError:
                        pass
                    else:
                        if isinstance(pd_type, type):
                            comp_types.append(pd_type)
                            try:
                                if isinstance(pd_type(), pd.core.dtypes.base.ExtensionDtype):
                                    comp_types.append(pd_type())
                            except TypeError:
                                pass
                    try:
                        pd_type = getattr(pd.core.dtypes.dtypes, type_)
                        if isinstance(pd_type, type):
                            comp_types.append(pd_type)
                    except AttributeError:
                        pass

                native_type = _native_type_type_map(type_)
                if native_type is not None:
                    comp_types.extend(native_type)

            # TODO: Remove when Numpy >=1.21 is pinned as a dependency
            _pandas_supports_extension_dtypes = version.parse(pd.__version__) >= version.parse(
                "0.24"
            )
            _numpy_doesnt_support_extensions_properly = version.parse(
                np.__version__
            ) < version.parse("1.21")
            if _numpy_doesnt_support_extensions_properly and _pandas_supports_extension_dtypes:
                # This works around a bug where Pandas nullable int types aren't compatible with Numpy dtypes  # noqa: E501 # FIXME CoP
                # Note: Can't do set difference, the whole bugfix is because numpy types can't be compared to  # noqa: E501 # FIXME CoP
                # ExtensionDtypes
                actual_type_is_ext_dtype = isinstance(
                    actual_column_type, pd.core.dtypes.base.ExtensionDtype
                )
                comp_types = {
                    dtype
                    for dtype in comp_types
                    if isinstance(dtype, pd.core.dtypes.base.ExtensionDtype)
                    == actual_type_is_ext_dtype
                }
            ###

            success = actual_column_type in comp_types

        return {
            "success": success,
            "result": {"observed_value": actual_column_type.type.__name__},
        }

    def _validate_sqlalchemy(self, actual_column_type, expected_types_list, execution_engine):
        if expected_types_list is None:
            success = True
        elif execution_engine.dialect_name in [
            GXSqlDialect.DATABRICKS,
            GXSqlDialect.POSTGRESQL,
            GXSqlDialect.SNOWFLAKE,
        ]:
            success = isinstance(actual_column_type, str) and any(
                actual_column_type.lower() == expected_type.lower()
                for expected_type in expected_types_list
            )
            return {
                "success": success,
                "result": {"observed_value": actual_column_type},
            }
        else:
            types = []
            for type_ in expected_types_list:
                types.extend(
                    _get_potential_sqlalchemy_types(
                        execution_engine=execution_engine, expected_type=type_
                    )
                )
            success = isinstance(actual_column_type, tuple(types))

        return {
            "success": success,
            "result": {"observed_value": type(actual_column_type).__name__},
        }

    def _validate_spark(
        self,
        actual_column_type,
        expected_types_list,
    ):
        if expected_types_list is None:
            success = True
        else:
            types = []
            for type_ in expected_types_list:
                try:
                    type_class = getattr(pyspark.types, type_)
                    types.append(type_class)
                except AttributeError:
                    logger.debug(f"Unrecognized type: {type_}")
            if len(types) == 0:
                raise ValueError("No recognized spark types in expected_types_list")  # noqa: TRY003 # FIXME CoP
            success = isinstance(actual_column_type, tuple(types))
        return {
            "success": success,
            "result": {"observed_value": type(actual_column_type).__name__},
        }

    @override
    def get_validation_dependencies(
        self,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> ValidationDependencies:
        from great_expectations.execution_engine import (
            PandasExecutionEngine,
        )

        # This calls BatchExpectation.get_validation_dependencies to set baseline validation_dependencies for the aggregate version  # noqa: E501 # FIXME CoP
        # of the expectation.
        # We need to keep this as super(ColumnMapExpectation, self), which calls
        # BatchExpectation.get_validation_dependencies instead of ColumnMapExpectation.get_validation_dependencies.  # noqa: E501 # FIXME CoP
        # This is because the map version of this expectation is only supported for Pandas, so we want the aggregate  # noqa: E501 # FIXME CoP
        # version for the other backends.
        validation_dependencies: ValidationDependencies = super(
            ColumnMapExpectation, self
        ).get_validation_dependencies(execution_engine, runtime_configuration)

        configuration = self.configuration

        # Only PandasExecutionEngine supports the column map version of the expectation.
        if isinstance(execution_engine, PandasExecutionEngine):
            column_name = configuration.kwargs.get("column") if configuration else None
            expected_types_list = configuration.kwargs.get("type_list") if configuration else None
            metric_kwargs = get_metric_kwargs(
                metric_name="table.column_types",
                configuration=configuration,
                runtime_configuration=runtime_configuration,
            )
            metric_domain_kwargs: dict = metric_kwargs.get("metric_domain_kwargs") or {}
            metric_value_kwargs = metric_kwargs.get("metric_value_kwargs") or {}
            table_column_types_configuration = MetricConfiguration(
                "table.column_types",
                metric_domain_kwargs=metric_domain_kwargs,
                metric_value_kwargs=metric_value_kwargs,
            )
            actual_column_types_list = execution_engine.resolve_metrics(
                [table_column_types_configuration]
            )[table_column_types_configuration.id]
            try:
                actual_column_type = [
                    type_dict["type"]
                    for type_dict in actual_column_types_list
                    if type_dict["name"] == column_name
                ][0]
            except IndexError:
                actual_column_type = None

            # only use column map version if column dtype is object
            if (
                actual_column_type
                and actual_column_type.type.__name__ == "object_"
                and expected_types_list is not None
            ):
                # this resets validation_dependencies using  ColumnMapExpectation.get_validation_dependencies  # noqa: E501 # FIXME CoP
                validation_dependencies = super().get_validation_dependencies(
                    execution_engine, runtime_configuration
                )

        # this adds table.column_types dependency for both aggregate and map versions of expectation
        column_types_metric_kwargs = get_metric_kwargs(
            metric_name="table.column_types",
            configuration=configuration,
            runtime_configuration=runtime_configuration,
        )
        validation_dependencies.set_metric_configuration(
            metric_name="table.column_types",
            metric_configuration=MetricConfiguration(
                metric_name="table.column_types",
                metric_domain_kwargs=column_types_metric_kwargs["metric_domain_kwargs"],
                metric_value_kwargs=column_types_metric_kwargs["metric_value_kwargs"],
            ),
        )

        return validation_dependencies

    @override
    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        from great_expectations.execution_engine import (
            PandasExecutionEngine,
            SparkDFExecutionEngine,
            SqlAlchemyExecutionEngine,
        )

        configuration = self.configuration
        column_name = configuration.kwargs.get("column")
        expected_types_list = configuration.kwargs.get("type_list")
        actual_column_types_list = metrics.get("table.column_types")
        actual_column_type = (
            [
                type_dict["type"]
                for type_dict in actual_column_types_list
                if type_dict["name"] == column_name
            ][0]
            if actual_column_types_list
            else []
        )

        if isinstance(execution_engine, PandasExecutionEngine):
            # only PandasExecutionEngine supports map version of expectation and
            # only when column type is object
            if actual_column_type.type.__name__ == "object_" and expected_types_list is not None:
                # this calls ColumnMapMetric._validate
                return super()._validate(metrics, runtime_configuration, execution_engine)
            return self._validate_pandas(
                actual_column_type=actual_column_type,
                expected_types_list=expected_types_list,
            )
        elif isinstance(execution_engine, SqlAlchemyExecutionEngine):
            return self._validate_sqlalchemy(
                actual_column_type=actual_column_type,
                expected_types_list=expected_types_list,
                execution_engine=execution_engine,
            )
        elif isinstance(execution_engine, SparkDFExecutionEngine):
            return self._validate_spark(
                actual_column_type=actual_column_type,
                expected_types_list=expected_types_list,
            )



from logging import Logger
from typing import Any, ClassVar, Literal, Optional, Type

from great_expectations.datasource.fluent import BatchMetadata, _SparkFilePathDatasource
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import CSVAsset
from great_expectations.datasource.fluent.data_connector import (
    S3DataConnector,
)

logger: Logger

class SparkS3Datasource(_SparkFilePathDatasource):
    # class attributes
    data_connector_type: ClassVar[Type[S3DataConnector]] = ...

    # instance attributes
    type: Literal["spark_s3"] = "spark_s3"

    # S3 specific attributes
    bucket: str
    boto3_options: dict[str, ConfigStr | Any] = {}
    def add_csv_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        s3_prefix: str = "",
        s3_delimiter: str = "/",
        s3_max_keys: int = 1000,
        s3_recursive_file_discovery: bool = False,
        header: bool = ...,
        infer_schema: bool = ...,
    ) -> CSVAsset: ...



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any

logger = logging.getLogger(__name__)

from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.core import RunIdentifier
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )
    from great_expectations.data_context.types.resource_identifiers import (
        ValidationResultIdentifier,
    )


class SlackRenderer(Renderer):
    def render(
        self,
        validation_result: ExpectationSuiteValidationResult,
        data_docs_pages: dict[ValidationResultIdentifier, dict[str, str]] | None = None,
        notify_with: list[str] | None = None,
        validation_result_urls: list[str] | None = None,
    ) -> list[dict]:
        data_docs_pages = data_docs_pages or {}
        notify_with = notify_with or []
        validation_result_urls = validation_result_urls or []
        blocks: list[dict] = []

        description_block = self._build_description_block(
            validation_result=validation_result,
            validation_result_urls=validation_result_urls,
        )
        blocks.append(description_block)

        for data_docs_page in data_docs_pages.values():
            report_element_block = self._build_report_element_block(
                data_docs_page=data_docs_page, notify_with=notify_with
            )
            if report_element_block:
                blocks.append(report_element_block)

        return blocks

    def _build_description_block(
        self,
        validation_result: ExpectationSuiteValidationResult,
        validation_result_urls: list[str],
    ) -> dict:
        validation_link = None
        summary_text = ""
        if validation_result_urls:
            if len(validation_result_urls) == 1:
                validation_link = validation_result_urls[0]

        n_checks_succeeded = validation_result.statistics["successful_expectations"]
        n_checks = validation_result.statistics["evaluated_expectations"]
        check_details_text = f"*{n_checks_succeeded}* of *{n_checks}* Expectations were met"

        expectation_suite_name = validation_result.suite_name
        data_asset_name = validation_result.asset_name or "__no_data_asset_name__"

        summary_text += f"\n*Asset*: `{data_asset_name}`  "
        # Slack does not allow links to local files due to security risks
        # DataDocs links will be added in a block after this summary text when applicable
        if validation_link and "file://" not in validation_link:
            summary_text += (
                f"\n*Expectation Suite*: {expectation_suite_name}  <{validation_link}|View Results>"
            )
        else:
            summary_text += f"\n*Expectation Suite*: `{expectation_suite_name}`"

        summary_text += f"\n*Summary*: {check_details_text}"

        return {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": summary_text,
            },
        }

    def concatenate_text_blocks(
        self,
        action_name: str,
        text_blocks: list[dict],
        success: bool,
        checkpoint_name: str,
        run_id: RunIdentifier,
    ) -> dict:
        all_blocks = [
            self._build_header(name=action_name, success=success, checkpoint_name=checkpoint_name)
        ]
        all_blocks.append(self._build_run_time_block(run_id=run_id))
        for block in text_blocks:
            all_blocks.append(block)
        all_blocks.append(self._build_divider())

        return {"blocks": all_blocks}

    def _build_header(self, name: str, success: bool, checkpoint_name: str) -> dict:
        status = "Success :white_check_mark:" if success else "Failure :no_entry:"
        return {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": f"{name} - {checkpoint_name} - {status}",
            },
        }

    def _build_run_time_block(self, run_id: RunIdentifier) -> dict:
        if run_id is not None:
            run_time = datetime.fromisoformat(str(run_id.run_time))
            formatted_run_time = run_time.strftime("%Y/%m/%d %I:%M %p")
        return {
            "type": "section",
            "text": {"type": "plain_text", "text": f"Runtime: {formatted_run_time}"},
        }

    def _build_divider(self) -> dict:
        return {"type": "divider"}

    def _build_footer(self) -> dict:
        documentation_url = "https://docs.greatexpectations.io/docs/terms/data_docs"
        return {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": f"Learn how to review validation results in Data Docs: {documentation_url}",  # noqa: E501 # FIXME CoP
                }
            ],
        }

    def _get_report_element(self, docs_link: str) -> dict[str, Any] | None:
        report_element = None
        if docs_link:
            try:
                # Slack does not allow links to local files due to security risks
                if "file://" in docs_link:
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*DataDocs* can be found here: `{docs_link}` \n (Please copy and paste link into "  # noqa: E501 # FIXME CoP
                            f"a browser to view)\n",
                        },
                    }
                else:
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*DataDocs* can be found here: <{docs_link}|{docs_link}>",
                        },
                    }
            except Exception as e:
                logger.warning(
                    f"""SlackRenderer had a problem with generating the docs link.
                    link used to generate the docs link is: {docs_link} and is of type: {type(docs_link)}.
                    Error: {e}"""  # noqa: E501 # FIXME CoP
                )
                return
        else:
            logger.warning("No docs link found. Skipping data docs link in Slack message.")
        return report_element

    def _build_report_element_block(
        self, data_docs_page: dict[str, str], notify_with: list[str]
    ) -> dict | None:
        if not data_docs_page:
            return None

        if notify_with:
            for docs_link_key in notify_with:
                if docs_link_key in data_docs_page:
                    docs_link = data_docs_page[docs_link_key]
                    report_element = self._get_report_element(docs_link)
                else:
                    logger.critical(
                        f"*ERROR*: Slack is trying to provide a link to the following DataDocs: `"
                        f"{docs_link_key!s}`, but it is not configured under `data_docs_sites` in the "  # noqa: E501 # FIXME CoP
                        f"`great_expectations.yml`\n"
                    )
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*ERROR*: Slack is trying to provide a link to the following DataDocs: "  # noqa: E501 # FIXME CoP
                            f"`{docs_link_key!s}`, but it is not configured under "
                            f"`data_docs_sites` in the `great_expectations.yml`\n",
                        },
                    }
                if report_element:
                    return report_element
        else:
            for docs_link_key, docs_link in data_docs_page.items():
                if docs_link_key == "class":
                    continue
                report_element = self._get_report_element(docs_link)
                return report_element

        return None



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, TypeVar

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import pydantic

# if we move this import into the TYPE_CHECKING block, we need to provide the
# Partitioner class when we update forward refs, so we just import here.
from great_expectations.core.freshness_diagnostics import (
    BatchDefinitionFreshnessDiagnostics,
)
from great_expectations.core.partitioners import ColumnPartitioner, FileNamePartitioner
from great_expectations.core.serdes import _EncodedValidationData, _IdentifierBundle
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.exceptions import (
    BatchDefinitionNotAddedError,
    BatchDefinitionNotFoundError,
    BatchDefinitionNotFreshError,
    DataAssetNotFoundError,
    DatasourceNotFoundError,
)

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.batch_request import (
        BatchParameters,
        BatchRequest,
    )
    from great_expectations.datasource.fluent.interfaces import Batch, DataAsset, Datasource

# Depending on the Asset
PartitionerT = TypeVar("PartitionerT", ColumnPartitioner, FileNamePartitioner, None)


@public_api
class BatchDefinition(pydantic.GenericModel, Generic[PartitionerT]):
    """Configuration for a batch of data.

    References the DataAsset to be used, and any additional parameters needed to fetch the data.
    """

    id: Optional[str] = None
    name: str
    partitioner: Optional[PartitionerT] = None

    # private attributes that must be set immediately after instantiation
    # Note that we're using type Any, but the getter setter ensure the right types.
    # If we actually specify DataAsset, pydantic errors out.
    _data_asset: Any = pydantic.PrivateAttr()

    @property
    @public_api
    def data_asset(self) -> DataAsset[Any, PartitionerT]:
        """
        The parent DataAsset for this Batch Definition.
        """
        return self._data_asset

    def set_data_asset(self, data_asset: DataAsset[Any, PartitionerT]) -> None:
        # pydantic prevents us from using @data_asset.setter
        self._data_asset = data_asset

    def build_batch_request(
        self, batch_parameters: Optional[BatchParameters] = None
    ) -> BatchRequest[PartitionerT]:
        """Build a BatchRequest from the asset and batch parameters."""
        return self.data_asset.build_batch_request(
            options=batch_parameters,
            partitioner=self.partitioner,
        )

    @public_api
    def save(self) -> None:
        """
        Save the batch definition to the underlying data context.
        """
        project_datasources = project_manager.get_datasources()
        data_source = self.data_asset.datasource
        project_datasources.set_datasource(name=data_source.name, ds=data_source)

    @public_api
    def get_batch(self, batch_parameters: Optional[BatchParameters] = None) -> Batch:
        """
        Retrieves a batch from the underlying asset. Defaults to the last batch
        from the asset's batch list.

        Args:
            batch_parameters: Additional parameters to be used in fetching the batch.

        Returns:
            A Batch of data.
        """
        batch_request = self.build_batch_request(batch_parameters=batch_parameters)
        return self.data_asset.get_batch(batch_request)

    @public_api
    def get_batch_identifiers_list(
        self, batch_parameters: Optional[BatchParameters] = None
    ) -> List[Dict]:
        """
        Retrieves a list of available batch identifiers.
        These identifiers can be used to fetch specific batches via batch_options.

        Args:
            batch_parameters: Additional parameters to be used in fetching the batch identifiers
                list.

        Returns:
            A list of batch identifiers.
        """
        batch_request = self.build_batch_request(batch_parameters=batch_parameters)
        return self.data_asset.get_batch_identifiers_list(batch_request)

    def is_fresh(self) -> BatchDefinitionFreshnessDiagnostics:
        diagnostics = self._is_added()
        if not diagnostics.success:
            return diagnostics
        return self._is_fresh()

    def _is_added(self) -> BatchDefinitionFreshnessDiagnostics:
        return BatchDefinitionFreshnessDiagnostics(
            errors=[] if self.id else [BatchDefinitionNotAddedError(name=self.name)]
        )

    def _is_fresh(self) -> BatchDefinitionFreshnessDiagnostics:
        datasource_dict = project_manager.get_datasources()

        datasource: Datasource | None
        try:
            datasource = datasource_dict[self.data_asset.datasource.name]
        except KeyError:
            datasource = None
        if not datasource:
            return BatchDefinitionFreshnessDiagnostics(
                errors=[
                    DatasourceNotFoundError(
                        f"Could not find datasource '{self.data_asset.datasource.name}'"
                    )
                ]
            )

        try:
            asset = datasource.get_asset(self.data_asset.name)
        except LookupError:
            asset = None
        if not asset:
            return BatchDefinitionFreshnessDiagnostics(
                errors=[DataAssetNotFoundError(f"Could not find asset '{self.data_asset.name}'")]
            )

        batch_def: BatchDefinition | None
        try:
            batch_def = asset.get_batch_definition(self.name)
        except KeyError:
            batch_def = None
        if not batch_def:
            return BatchDefinitionFreshnessDiagnostics(
                errors=[
                    BatchDefinitionNotFoundError(f"Could not find batch definition '{self.name}'")
                ]
            )

        return BatchDefinitionFreshnessDiagnostics(
            errors=[] if self == batch_def else [BatchDefinitionNotFreshError(name=self.name)]
        )

    def identifier_bundle(self) -> _EncodedValidationData:
        # Utilized as a custom json_encoder
        diagnostics = self.is_fresh()
        diagnostics.raise_for_error()

        asset = self.data_asset
        data_source = asset.datasource

        data_source_bundle = _IdentifierBundle(
            name=data_source.name,
            id=str(data_source.id) if data_source.id else None,
        )
        asset_bundle = _IdentifierBundle(
            name=asset.name,
            id=str(asset.id) if asset.id else None,
        )
        batch_definition_bundle = _IdentifierBundle(
            name=self.name,
            id=str(self.id) if self.id else None,
        )

        return _EncodedValidationData(
            datasource=data_source_bundle,
            asset=asset_bundle,
            batch_definition=batch_definition_bundle,
        )
